# ------------------------------------------------------------------------------
# This file is part of PyTango (http://pytango.rtfd.io)
#
# Copyright 2006-2012 CELLS / ALBA Synchrotron, Bellaterra, Spain
# Copyright 2013-2014 European Synchrotron Radiation Facility, Grenoble, France
#
# Distributed under the terms of the GNU Lesser General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
# ------------------------------------------------------------------------------

"""
This is an internal PyTango module.
"""

__all__ = ("base_types_init",)

__docformat__ = "restructuredtext"

from ._tango import (StdStringVector, StdLongVector, StdDoubleVector,
                     CommandInfoList, AttributeInfoList, AttributeInfoListEx, DbData,
                     DbDevInfos, DbDevExportInfos, DbDevImportInfos, DbHistoryList,
                     DeviceDataHistoryList, StdGroupReplyVector,
                     StdGroupCmdReplyVector, StdGroupAttrReplyVector,
                     ArchiveEventInfo, EventData, AttrConfEventData, AttributeAlarmInfo,
                     AttributeDimension, AttributeEventInfo, DeviceAttributeConfig,
                     AttributeInfo, AttributeInfoEx, ChangeEventInfo, PeriodicEventInfo,
                     DevCommandInfo, CommandInfo, DataReadyEventData, DeviceInfo,
                     LockerInfo, PollDevice, TimeVal, AttrWriteType, AttrDataFormat, DispLevel)

from .utils import document_method, is_integer
from .utils import document_enum as __document_enum
from .utils import seq_2_StdStringVector, StdStringVector_2_seq


def __StdVector__add(self, seq):
    ret = seq.__class__(self)
    ret.extend(seq)
    return ret


def __StdVector__mul(self, n):
    ret = self.__class__()
    for _ in range(n):
        ret.extend(self)
    return ret


def __StdVector__imul(self, n):
    ret = self.__class__()
    for _ in range(n):
        ret.extend(self)
    return ret


def __StdVector__getitem(self, key):
    if is_integer(key) or key.step is None:
        return self.__original_getitem(key)

    res = self.__class__()
    nb = len(self)
    start = key.start or 0
    stop = key.stop or nb
    if start >= nb:
        return res
    if stop > nb:
        stop = nb

    for i in range(start, stop, key.step or 1):
        res.append(self[i])

    return res


def __fillVectorClass(klass):
    klass.__add__ = __StdVector__add
    klass.__mul__ = __StdVector__mul
    klass.__imul__ = __StdVector__imul
    klass.__original_getitem = klass.__getitem__
    klass.__getitem__ = __StdVector__getitem


# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
# DeviceAttributeConfig pickle
# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~

def __DeviceAttributeConfig__getinitargs__(self):
    return ()


def __DeviceAttributeConfig__getstate__(self):
    return (
        self.name,
        int(self.writable),
        int(self.data_format),
        self.data_type,
        self.max_dim_x,
        self.max_dim_y,
        self.description,
        self.label,
        self.unit,
        self.standard_unit,
        self.display_unit,
        self.format,
        self.min_value,
        self.max_value,
        self.min_alarm,
        self.max_alarm,
        self.writable_attr_name,
        StdStringVector_2_seq(self.extensions))


def __DeviceAttributeConfig__setstate__(self, state):
    self.name = state[0]
    self.writable = AttrWriteType(state[1])
    self.data_format = AttrDataFormat(state[2])
    self.data_type = state[3]
    self.max_dim_x = state[4]
    self.max_dim_y = state[5]
    self.description = state[6]
    self.label = state[7]
    self.unit = state[8]
    self.standard_unit = state[9]
    self.display_unit = state[10]
    self.format = state[11]
    self.min_value = state[12]
    self.max_value = state[13]
    self.min_alarm = state[14]
    self.max_alarm = state[15]
    self.writable_attr_name = state[16]
    self.extensions = seq_2_StdStringVector(state[17])


# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
# AttributeInfo pickle
# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~

def __AttributeInfo__getinitargs__(self):
    return ()


def __AttributeInfo__getstate__(self):
    ret = list(__DeviceAttributeConfig__getstate__(self))
    ret.append(int(self.disp_level))
    return tuple(ret)


def __AttributeInfo__setstate__(self, state):
    __DeviceAttributeConfig__setstate__(self, state)
    self.disp_level = DispLevel(state[18])


# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
# AttributeAlarmInfo pickle
# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~

def __AttributeAlarmInfo__getinitargs__(self):
    return ()


def __AttributeAlarmInfo__getstate__(self):
    return (
        self.min_alarm,
        self.max_alarm,
        self.min_warning,
        self.max_warning,
        self.delta_t,
        self.delta_val,
        StdStringVector_2_seq(self.extensions))


def __AttributeAlarmInfo__setstate__(self, state):
    self.min_alarm = state[0]
    self.max_alarm = state[1]
    self.min_warning = state[2]
    self.max_warning = state[3]
    self.delta_t = state[4]
    self.delta_val = state[5]
    self.extensions = seq_2_StdStringVector(state[6])


# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
# ChangeEventInfo pickle
# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~

def __ChangeEventInfo__getinitargs__(self):
    return ()


def __ChangeEventInfo__getstate__(self):
    return (
        self.rel_change,
        self.abs_change,
        StdStringVector_2_seq(self.extensions))


def __ChangeEventInfo__setstate__(self, state):
    self.rel_change = state[0]
    self.abs_change = state[1]
    self.extensions = seq_2_StdStringVector(state[2])


# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
# PeriodicEventInfo pickle
# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~

def __PeriodicEventInfo__getinitargs__(self):
    return ()


def __PeriodicEventInfo__getstate__(self):
    return (
        self.period,
        StdStringVector_2_seq(self.extensions))


def __PeriodicEventInfo__setstate__(self, state):
    self.period = state[0]
    self.extensions = seq_2_StdStringVector(state[1])


# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
# ArchiveEventInfo pickle
# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~

def __ArchiveEventInfo__getinitargs__(self):
    return ()


def __ArchiveEventInfo__getstate__(self):
    return (
        self.archive_rel_change,
        self.archive_abs_change,
        self.archive_period,
        StdStringVector_2_seq(self.extensions))


def __ArchiveEventInfo__setstate__(self, state):
    self.archive_rel_change = state[0]
    self.archive_abs_change = state[1]
    self.archive_period = state[2]
    self.extensions = seq_2_StdStringVector(state[3])


# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
# AttributeEventInfo pickle
# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~

def __AttributeEventInfo__getinitargs__(self):
    return ()


def __AttributeEventInfo__getstate__(self):
    return (
        self.ch_event,
        self.per_event,
        self.arch_event)


def __AttributeEventInfo__setstate__(self, state):
    self.ch_event = state[0]
    self.per_event = state[1]
    self.arch_event = state[2]


# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
# AttributeInfoEx pickle
# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~

def __AttributeInfoEx__getinitargs__(self):
    return ()


def __AttributeInfoEx__getstate__(self):
    ret = list(__AttributeInfo__getstate__(self))
    ret.append(self.alarms)
    ret.append(self.events)
    ret.append(StdStringVector_2_seq(self.sys_extensions))
    return tuple(ret)


def __AttributeInfoEx__setstate__(self, state):
    __AttributeInfo__setstate__(self, state)
    self.alarms = state[19]
    self.events = state[20]
    self.sys_extensions = seq_2_StdStringVector(state[21])


def __init_base_types():
    v_klasses = (
        StdStringVector, StdLongVector, StdDoubleVector, CommandInfoList,
        AttributeInfoList, AttributeInfoListEx, DbData, DbDevInfos,
        DbDevExportInfos, DbDevImportInfos, DbHistoryList,
        DeviceDataHistoryList, StdGroupReplyVector,
        StdGroupCmdReplyVector, StdGroupAttrReplyVector)

    for v_klass in v_klasses:
        __fillVectorClass(v_klass)

    DeviceAttributeConfig.__getinitargs__ = __DeviceAttributeConfig__getinitargs__
    DeviceAttributeConfig.__getstate__ = __DeviceAttributeConfig__getstate__
    DeviceAttributeConfig.__setstate__ = __DeviceAttributeConfig__setstate__

    AttributeInfo.__getinitargs__ = __AttributeInfo__getinitargs__
    AttributeInfo.__getstate__ = __AttributeInfo__getstate__
    AttributeInfo.__setstate__ = __AttributeInfo__setstate__

    AttributeAlarmInfo.__getinitargs__ = __AttributeAlarmInfo__getinitargs__
    AttributeAlarmInfo.__getstate__ = __AttributeAlarmInfo__getstate__
    AttributeAlarmInfo.__setstate__ = __AttributeAlarmInfo__setstate__

    ChangeEventInfo.__getinitargs__ = __ChangeEventInfo__getinitargs__
    ChangeEventInfo.__getstate__ = __ChangeEventInfo__getstate__
    ChangeEventInfo.__setstate__ = __ChangeEventInfo__setstate__

    PeriodicEventInfo.__getinitargs__ = __PeriodicEventInfo__getinitargs__
    PeriodicEventInfo.__getstate__ = __PeriodicEventInfo__getstate__
    PeriodicEventInfo.__setstate__ = __PeriodicEventInfo__setstate__

    ArchiveEventInfo.__getinitargs__ = __ArchiveEventInfo__getinitargs__
    ArchiveEventInfo.__getstate__ = __ArchiveEventInfo__getstate__
    ArchiveEventInfo.__setstate__ = __ArchiveEventInfo__setstate__

    AttributeEventInfo.__getinitargs__ = __AttributeEventInfo__getinitargs__
    AttributeEventInfo.__getstate__ = __AttributeEventInfo__getstate__
    AttributeEventInfo.__setstate__ = __AttributeEventInfo__setstate__

    AttributeInfoEx.__getinitargs__ = __AttributeInfoEx__getinitargs__
    AttributeInfoEx.__getstate__ = __AttributeInfoEx__getstate__
    AttributeInfoEx.__setstate__ = __AttributeInfoEx__setstate__


def __doc_base_types():
    def document_enum(enum_name, desc):
        import tango
        __document_enum(tango, enum_name, desc)

    document_enum("ExtractAs", """
    Defines what will go into value field of DeviceAttribute, or what will
    Attribute.get_write_value() return... Not all the possible values are valid
    in all the cases.

    Valid possible values are:

        - Numpy    : Value will be stored in [value, w_value]. If the
          attribute is an scalar, they will contain a value. If it's
          an SPECTRUM or IMAGE it will be exported as a numpy array.
        - Tuple    : Value will be stored in [value, w_value]. If the
          attribute is an scalar, they will contain a value. If it's
          an SPECTRUM or IMAGE it will be exported as a tuple or
          tuple of tuples.
        - List     : Value will be stored in [value, w_value]. If the
          attribute is an scalar, they will contain a value. If it's
          an SPECTRUM or IMAGE it will be exported as a list or list
          of lists
        - String   : The data will be stored 'as is', the binary data
          as it comes from TangoC++ in 'value'.
        - Nothing  : The value will not be extracted from DeviceAttribute
    """)

    document_enum("CmdArgType", """
    An enumeration representing the command argument type.

        - DevVoid
        - DevBoolean
        - DevShort
        - DevLong
        - DevFloat
        - DevDouble
        - DevUShort
        - DevULong
        - DevString
        - DevVarCharArray
        - DevVarShortArray
        - DevVarLongArray
        - DevVarFloatArray
        - DevVarDoubleArray
        - DevVarUShortArray
        - DevVarULongArray
        - DevVarStringArray
        - DevVarLongStringArray
        - DevVarDoubleStringArray
        - DevState
        - ConstDevString
        - DevVarBooleanArray
        - DevUChar
        - DevLong64
        - DevULong64
        - DevVarLong64Array
        - DevVarULong64Array
        - DevInt
        - DevEncoded
        - DevEnum
        - DevPipeBlob
    """)

    document_enum("LockerLanguage", """
    An enumeration representing the programming language in which the
    client application who locked is written.

        - CPP : C++/Python language
        - JAVA : Java language

    New in PyTango 7.0.0
    """)

    document_enum("MessBoxType", """
    An enumeration representing the MessBoxType

        - STOP
        - INFO

    New in PyTango 7.0.0
    """)

    document_enum("PollObjType", """
    An enumeration representing the PollObjType

        - POLL_CMD
        - POLL_ATTR
        - EVENT_HEARTBEAT
        - STORE_SUBDEV

    New in PyTango 7.0.0
    """)

    document_enum("PollCmdCode", """
    An enumeration representing the PollCmdCode

        - POLL_ADD_OBJ
        - POLL_REM_OBJ
        - POLL_START
        - POLL_STOP
        - POLL_UPD_PERIOD
        - POLL_REM_DEV
        - POLL_EXIT
        - POLL_REM_EXT_TRIG_OBJ
        - POLL_ADD_HEARTBEAT
        - POLL_REM_HEARTBEAT

    New in PyTango 7.0.0
    """)

    document_enum("SerialModel", """
    An enumeration representing the type of serialization performed by the device server

        - BY_DEVICE
        - BY_CLASS
        - BY_PROCESS
        - NO_SYNC
    """)

    document_enum("AttReqType", """
    An enumeration representing the type of attribute request

        - READ_REQ
        - WRITE_REQ
    """)

    document_enum("LockCmdCode", """
    An enumeration representing the LockCmdCode

        - LOCK_ADD_DEV
        - LOCK_REM_DEV
        - LOCK_UNLOCK_ALL_EXIT
        - LOCK_EXIT

    New in PyTango 7.0.0
    """)

    document_enum("LogLevel", """
    An enumeration representing the LogLevel

        - LOG_OFF
        - LOG_FATAL
        - LOG_ERROR
        - LOG_WARN
        - LOG_INFO
        - LOG_DEBUG

    New in PyTango 7.0.0
    """)

    document_enum("LogTarget", """
    An enumeration representing the LogTarget

        - LOG_CONSOLE
        - LOG_FILE
        - LOG_DEVICE

    New in PyTango 7.0.0
    """)

    document_enum("EventType", """
    An enumeration representing event type

        - CHANGE_EVENT
        - QUALITY_EVENT
        - PERIODIC_EVENT
        - ARCHIVE_EVENT
        - USER_EVENT
        - ATTR_CONF_EVENT
        - DATA_READY_EVENT
        - INTERFACE_CHANGE_EVENT
        - PIPE_EVENT

        *DATA_READY_EVENT - New in PyTango 7.0.0*
        *INTERFACE_CHANGE_EVENT - New in PyTango 9.2.2*
        *PIPE_EVENT - New in PyTango 9.2.2*

    """)

    document_enum("AttrSerialModel", """
    An enumeration representing the AttrSerialModel

        - ATTR_NO_SYNC
        - ATTR_BY_KERNEL
        - ATTR_BY_USER

    New in PyTango 7.1.0
    """)

    document_enum("KeepAliveCmdCode", """
    An enumeration representing the KeepAliveCmdCode

        - EXIT_TH

    New in PyTango 7.0.0
    """)

    document_enum("AccessControlType", """
    An enumeration representing the AccessControlType

        - ACCESS_READ
        - ACCESS_WRITE

    New in PyTango 7.0.0
    """)

    document_enum("asyn_req_type", """
    An enumeration representing the asynchronous request type

        - POLLING
        - CALLBACK
        - ALL_ASYNCH
    """)

    document_enum("cb_sub_model", """
    An enumeration representing callback sub model

        - PUSH_CALLBACK
        - PULL_CALLBACK
    """)

    document_enum("AttrQuality", """
    An enumeration representing the attribute quality

        - ATTR_VALID
        - ATTR_INVALID
        - ATTR_ALARM
        - ATTR_CHANGING
        - ATTR_WARNING
    """)

    document_enum("AttrWriteType", """
    An enumeration representing the attribute type

        - READ
        - READ_WITH_WRITE
        - WRITE
        - READ_WRITE
    """)

    document_enum("AttrDataFormat", """
    An enumeration representing the attribute format

        - SCALAR
        - SPECTRUM
        - IMAGE
        - FMT_UNKNOWN
    """)

    document_enum("PipeWriteType", """
    An enumeration representing the pipe type

        - PIPE_READ
        - PIPE_READ_WRITE
    """)

    document_enum("DevSource", """
    An enumeration representing the device source for data

        - DEV
        - CACHE
        - CACHE_DEV
    """)

    document_enum("ErrSeverity", """
    An enumeration representing the error severity

        - WARN
        - ERR
        - PANIC
    """)

    document_enum("DevState", """
    An enumeration representing the device state

        - ON
        - OFF
        - CLOSE
        - OPEN
        - INSERT
        - EXTRACT
        - MOVING
        - STANDBY
        - FAULT
        - INIT
        - RUNNING
        - ALARM
        - DISABLE
        - UNKNOWN
    """)

    document_enum("DispLevel", """
    An enumeration representing the display level

        - OPERATOR
        - EXPERT
    """)

    document_enum("GreenMode", """
    An enumeration representing the GreenMode

        - Synchronous
        - Futures
        - Gevent

    New in PyTango 8.1.0
    """)

    ArchiveEventInfo.__doc__ = """
    A structure containing available archiving event information for an attribute
    with the folowing members:

        - archive_rel_change : (str) relative change that will generate an event
        - archive_abs_change : (str) absolute change that will generate an event
        - archive_period : (str) archive period
        - extensions : (sequence<str>) extensions (currently not used)"""

    EventData.__doc__ = """
    This class is used to pass data to the callback method when an event
    is sent to the client. It contains the following public fields:

         - device : (DeviceProxy) The DeviceProxy object on which the call was
           executed.
         - attr_name : (str) The attribute name
         - event : (str) The event name
         - attr_value : (DeviceAttribute) The attribute data (DeviceAttribute)
         - err : (bool) A boolean flag set to true if the request failed. False
           otherwise
         - errors : (sequence<DevError>) The error stack
         - reception_date: (TimeVal)
    """

    document_method(EventData, "get_date", """
    get_date(self) -> TimeVal

            Returns the timestamp of the event.

        Parameters : None
        Return     : (TimeVal) the timestamp of the event

        New in PyTango 7.0.0
    """)

    AttrConfEventData.__doc__ = """
    This class is used to pass data to the callback method when a
    configuration event is sent to the client. It contains the
    following public fields:

        - device : (DeviceProxy) The DeviceProxy object on which the call was executed
        - attr_name : (str) The attribute name
        - event : (str) The event name
        - attr_conf : (AttributeInfoEx) The attribute data
        - err : (bool) A boolean flag set to true if the request failed. False
          otherwise
        - errors : (sequence<DevError>) The error stack
        - reception_date: (TimeVal)
    """

    document_method(AttrConfEventData, "get_date", """
    get_date(self) -> TimeVal

            Returns the timestamp of the event.

        Parameters : None
        Return     : (TimeVal) the timestamp of the event

        New in PyTango 7.0.0
    """)

    AttributeAlarmInfo.__doc__ = """
    A structure containing available alarm information for an attribute
    with the folowing members:

        - min_alarm : (str) low alarm level
        - max_alarm : (str) high alarm level
        - min_warning : (str) low warning level
        - max_warning : (str) high warning level
        - delta_t : (str) time delta
        - delta_val : (str) value delta
        - extensions : (StdStringVector) extensions (currently not used)"""

    AttributeDimension.__doc__ = """
    A structure containing x and y attribute data dimensions with
    the following members:

        - dim_x : (int) x dimension
        - dim_y : (int) y dimension"""

    AttributeEventInfo.__doc__ = """
    A structure containing available event information for an attribute
    with the folowing members:

        - ch_event : (ChangeEventInfo) change event information
        - per_event : (PeriodicEventInfo) periodic event information
        - arch_event :  (ArchiveEventInfo) archiving event information"""

    DeviceAttributeConfig.__doc__ = """
    A base structure containing available information for an attribute
    with the following members:

        - name : (str) attribute name
        - writable : (AttrWriteType) write type (R, W, RW, R with W)
        - data_format : (AttrDataFormat) data format (SCALAR, SPECTRUM, IMAGE)
        - data_type : (int) attribute type (float, string,..)
        - max_dim_x : (int) first dimension of attribute (spectrum or image attributes)
        - max_dim_y : (int) second dimension of attribute(image attribute)
        - description : (int) attribute description
        - label : (str) attribute label (Voltage, time, ...)
        - unit : (str) attribute unit (V, ms, ...)
        - standard_unit : (str) standard unit
        - display_unit : (str) display unit
        - format : (str) how to display the attribute value (ex: for floats could be '%6.2f')
        - min_value : (str) minimum allowed value
        - max_value : (str) maximum allowed value
        - min_alarm : (str) low alarm level
        - max_alarm : (str) high alarm level
        - writable_attr_name : (str) name of the writable attribute
        - extensions : (StdStringVector) extensions (currently not used)"""

    AttributeInfo.__doc__ = """
    A structure (inheriting from :class:`DeviceAttributeConfig`) containing
    available information for an attribute with the following members:

        - disp_level : (DispLevel) display level (OPERATOR, EXPERT)

        Inherited members are:

            - name : (str) attribute name
            - writable : (AttrWriteType) write type (R, W, RW, R with W)
            - data_format : (AttrDataFormat) data format (SCALAR, SPECTRUM, IMAGE)
            - data_type : (int) attribute type (float, string,..)
            - max_dim_x : (int) first dimension of attribute (spectrum or image attributes)
            - max_dim_y : (int) second dimension of attribute(image attribute)
            - description : (int) attribute description
            - label : (str) attribute label (Voltage, time, ...)
            - unit : (str) attribute unit (V, ms, ...)
            - standard_unit : (str) standard unit
            - display_unit : (str) display unit
            - format : (str) how to display the attribute value (ex: for floats could be '%6.2f')
            - min_value : (str) minimum allowed value
            - max_value : (str) maximum allowed value
            - min_alarm : (str) low alarm level
            - max_alarm : (str) high alarm level
            - writable_attr_name : (str) name of the writable attribute
            - extensions : (StdStringVector) extensions (currently not used)"""

    AttributeInfoEx.__doc__ = """
    A structure (inheriting from :class:`AttributeInfo`) containing
    available information for an attribute with the following members:

        - alarms : object containing alarm information (see AttributeAlarmInfo).
        - events : object containing event information (see AttributeEventInfo).
        - sys_extensions : StdStringVector

        Inherited members are:

            - name : (str) attribute name
            - writable : (AttrWriteType) write type (R, W, RW, R with W)
            - data_format : (AttrDataFormat) data format (SCALAR, SPECTRUM, IMAGE)
            - data_type : (int) attribute type (float, string,..)
            - max_dim_x : (int) first dimension of attribute (spectrum or image attributes)
            - max_dim_y : (int) second dimension of attribute(image attribute)
            - description : (int) attribute description
            - label : (str) attribute label (Voltage, time, ...)
            - unit : (str) attribute unit (V, ms, ...)
            - standard_unit : (str) standard unit
            - display_unit : (str) display unit
            - format : (str) how to display the attribute value (ex: for floats could be '%6.2f')
            - min_value : (str) minimum allowed value
            - max_value : (str) maximum allowed value
            - min_alarm : (str) low alarm level
            - max_alarm : (str) high alarm level
            - writable_attr_name : (str) name of the writable attribute
            - extensions : (StdStringVector) extensions (currently not used)
            - disp_level : (DispLevel) display level (OPERATOR, EXPERT)"""

    AttributeInfoList.__doc__ = """
    List of AttributeInfo objects, containing available information for the attributes"""

    AttributeInfoListEx.__doc__ = """
    List of AttributeInfoEx objects, containing available information for the attributes"""

    ChangeEventInfo.__doc__ = """
    A structure containing available change event information for an attribute
    with the folowing members:

        - rel_change : (str) relative change that will generate an event
        - abs_change : (str) absolute change that will generate an event
        - extensions : (StdStringVector) extensions (currently not used)"""

    PeriodicEventInfo.__doc__ = """
    A structure containing available periodic event information for an attribute
    with the folowing members:

        - period : (str) event period
        - extensions : (StdStringVector) extensions (currently not used)"""

    DevCommandInfo.__doc__ = """
    A device command info with the following members:

        - cmd_name : (str) command name
        - cmd_tag : command as binary value (for TACO)
        - in_type : (CmdArgType) input type
        - out_type : (CmdArgType) output type
        - in_type_desc : (str) description of input type
        - out_type_desc : (str) description of output type

    New in PyTango 7.0.0"""

    CommandInfo.__doc__ = """
    A device command info (inheriting from :class:`DevCommandInfo`) with the following members:

        - disp_level : (DispLevel) command display level

        Inherited members are (from :class:`DevCommandInfo`):

            - cmd_name : (str) command name
            - cmd_tag : (str) command as binary value (for TACO)
            - in_type : (CmdArgType) input type
            - out_type : (CmdArgType) output type
            - in_type_desc : (str) description of input type
            - out_type_desc : (str) description of output type"""

    DataReadyEventData.__doc__ = """
    This class is used to pass data to the callback method when an
    attribute data ready event is sent to the clien. It contains the
    following public fields:

        - device : (DeviceProxy) The DeviceProxy object on which the call was executed
        - attr_name : (str) The attribute name
        - event : (str) The event name
        - attr_data_type : (int) The attribute data type
        - ctr : (int) The user counter. Set to 0 if not defined when sent by the
          server
        - err : (bool) A boolean flag set to true if the request failed. False
          otherwise
        - errors : (sequence<DevError>) The error stack
        - reception_date: (TimeVal)

        New in PyTango 7.0.0"""

    DeviceInfo.__doc__ = """
    A structure containing available information for a device with the"
    following members:

        - dev_class : (str) device class
        - server_id : (str) server ID
        - server_host : (str) host name
        - server_version : (str) server version
        - doc_url : (str) document url"""

    LockerInfo.__doc__ = """
    A structure with information about the locker with the folowing members:

        - ll : (tango.LockerLanguage) the locker language
        - li : (pid_t / UUID) the locker id
        - locker_host : (str) the host
        - locker_class : (str) the class

        pid_t should be an int, UUID should be a tuple of four numbers.

        New in PyTango 7.0.0"""

    PollDevice.__doc__ = """
    A structure containing PollDevice information with the folowing members:

        - dev_name : (str) device name
        - ind_list : (sequence<int>) index list

        New in PyTango 7.0.0"""

    document_method(DataReadyEventData, "get_date", """
    get_date(self) -> TimeVal

            Returns the timestamp of the event.

        Parameters : None
        Return     : (TimeVal) the timestamp of the event

        New in PyTango 7.0.0
    """)

    TimeVal.__doc__ = """
    Time value structure with the following members:

        - tv_sec : seconds
        - tv_usec : microseconds
        - tv_nsec : nanoseconds"""


def base_types_init(doc=True):
    __init_base_types()
    if doc:
        __doc_base_types()



import os

import pytest

import tango

from tango import AttrWriteType, GreenMode
from tango.asyncio_executor import AsyncioExecutor
from tango.device_server import get_worker
from tango.gevent_executor import GeventExecutor
from tango.green import SynchronousExecutor
from tango.server import Device
from tango.server import command, attribute, device_property
from tango.test_utils import (
    DeviceTestContext,
    MultiDeviceTestContext,
    SimpleDevice,
    ClassicAPISimpleDeviceImpl,
    ClassicAPISimpleDeviceClass
)


WINDOWS = "nt" in os.name


class Device1(Device):

    def init_device(self):
        super(Device, self).init_device()
        self.set_state(tango.DevState.ON)
        self._attr1 = 100

    @attribute
    def attr1(self):
        return self._attr1


class Device2(Device):

    def init_device(self):
        super(Device, self).init_device()
        self.set_state(tango.DevState.ON)
        self._attr2 = 200

    @attribute
    def attr2(self):
        return self._attr2


class Device1GreenModeUnspecified(Device1):
    pass


class Device1Synchronous(Device1):
    green_mode = GreenMode.Synchronous


class Device1Gevent(Device1):
    green_mode = GreenMode.Gevent


class Device1Asyncio(Device1):
    green_mode = GreenMode.Asyncio


class Device2GreenModeUnspecified(Device2):
    pass


class Device2Synchronous(Device2):
    green_mode = GreenMode.Synchronous


class Device2Gevent(Device2):
    green_mode = GreenMode.Gevent


class Device2Asyncio(Device2):
    green_mode = GreenMode.Asyncio


class Device1AsyncInit(Device1):
    green_mode = GreenMode.Asyncio

    async def init_device(self):
        await super().init_device()
        self._attr1 = 150


def test_single_device(server_green_mode):
    class TestDevice(Device1):
        green_mode = server_green_mode

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.attr1 == 100


def test_single_async_init_device():
    with DeviceTestContext(Device1AsyncInit) as proxy:
        assert proxy.attr1 == 150


def test_single_device_old_api():
    with DeviceTestContext(ClassicAPISimpleDeviceImpl, ClassicAPISimpleDeviceClass) as proxy:
        assert proxy.attr1 == 100


@pytest.mark.parametrize(
    "class_field, device",
    [
        (SimpleDevice, SimpleDevice),
        ("tango.test_utils.SimpleDevice", SimpleDevice),
        (
            ("tango.test_utils.ClassicAPISimpleDeviceClass", "tango.test_utils.ClassicAPISimpleDeviceImpl"),
            ClassicAPISimpleDeviceImpl
        ),
        (
            ("tango.test_utils.ClassicAPISimpleDeviceClass", ClassicAPISimpleDeviceImpl),
            ClassicAPISimpleDeviceImpl
        ),
        (
            (ClassicAPISimpleDeviceClass, "tango.test_utils.ClassicAPISimpleDeviceImpl"),
            ClassicAPISimpleDeviceImpl
        ),
        (
            (ClassicAPISimpleDeviceClass, ClassicAPISimpleDeviceImpl),
            ClassicAPISimpleDeviceImpl
        ),
    ]
)
def test_multi_devices_info(class_field, device):
    devices_info = ({"class": class_field, "devices": [{"name": "test/device1/1"}]},)

    dev_class = device if isinstance(device, str) else device.__name__

    with MultiDeviceTestContext(devices_info) as context:
        proxy1 = context.get_device("test/device1/1")
        assert proxy1.info().dev_class == dev_class


def test_multi_with_single_device(server_green_mode):
    class TestDevice(Device1):
        green_mode = server_green_mode

    devices_info = ({"class": TestDevice, "devices": [{"name": "test/device1/1"}]},)

    with MultiDeviceTestContext(devices_info) as context:
        proxy1 = context.get_device("test/device1/1")
        assert proxy1.attr1 == 100


def test_multi_with_single_device_old_api():
    devices_info = (
        {
            "class": (ClassicAPISimpleDeviceClass, ClassicAPISimpleDeviceImpl),
            "devices": [{"name": "test/device1/1"}],
        },
    )

    with MultiDeviceTestContext(devices_info) as context:
        proxy1 = context.get_device("test/device1/1")
        assert proxy1.attr1 == 100


def test_multi_with_two_devices(server_green_mode):
    class TestDevice1(Device1):
        green_mode = server_green_mode

    class TestDevice2(Device2):
        green_mode = server_green_mode

    devices_info = (
        {"class": TestDevice1, "devices": [{"name": "test/device1/1"}]},
        {"class": TestDevice2, "devices": [{"name": "test/device2/1"}]},
    )

    with MultiDeviceTestContext(devices_info) as context:
        proxy1 = context.get_device("test/device1/1")
        proxy2 = context.get_device("test/device2/1")
        assert proxy1.State() == tango.DevState.ON
        assert proxy2.State() == tango.DevState.ON
        assert proxy1.attr1 == 100
        assert proxy2.attr2 == 200


@pytest.mark.parametrize(
    "first_type, second_type, exception_type",
    [
        (Device1GreenModeUnspecified, Device2GreenModeUnspecified, None),
        (Device1GreenModeUnspecified, Device2Synchronous, None),
        (Device1GreenModeUnspecified, Device2Gevent, ValueError),
        (Device1GreenModeUnspecified, Device2Asyncio, ValueError),
        (Device1Synchronous, Device2GreenModeUnspecified, None),
        (Device1Synchronous, Device2Synchronous, None),
        (Device1Synchronous, Device2Gevent, ValueError),
        (Device1Synchronous, Device2Asyncio, ValueError),
        (Device1Asyncio, Device2GreenModeUnspecified, ValueError),
        (Device1Asyncio, Device2Synchronous, ValueError),
        (Device1Asyncio, Device2Gevent, ValueError),
        (Device1Asyncio, Device2Asyncio, None),
        (Device1Gevent, Device2GreenModeUnspecified, ValueError),
        (Device1Gevent, Device2Synchronous, ValueError),
        (Device1Gevent, Device2Gevent, None),
        (Device1Gevent, Device2Asyncio, ValueError),
    ]
)
def test_multi_with_mixed_device_green_modes(first_type, second_type, exception_type):

    devices_info = (
        {"class": first_type, "devices": [{"name": "test/device1/1"}]},
        {"class": second_type, "devices": [{"name": "test/device2/1"}]},
    )

    if exception_type is None:
        with MultiDeviceTestContext(devices_info):
            pass
    else:
        with pytest.raises(exception_type, match=r"mixed green mode"):
            with MultiDeviceTestContext(devices_info):
                pass


@pytest.mark.parametrize(
    "device_type, green_mode, global_mode, exception_type, executor_type",
    [
        # If a device specifies its green mode explicitly, then both
        # green_mode kwarg and global green mode are ignored. The device must use its specified mode.
        (Device1Synchronous, GreenMode.Asyncio, GreenMode.Asyncio, None, SynchronousExecutor),
        (Device1Synchronous, GreenMode.Gevent, GreenMode.Gevent, None, SynchronousExecutor),
        (Device1Asyncio, GreenMode.Synchronous, GreenMode.Synchronous, None, AsyncioExecutor),
        (Device1Asyncio, GreenMode.Gevent, GreenMode.Gevent, None, AsyncioExecutor),
        (Device1Gevent, GreenMode.Synchronous, GreenMode.Synchronous, None, GeventExecutor),
        (Device1Gevent, GreenMode.Asyncio, GreenMode.Asyncio, None, GeventExecutor),

        # If device doesn't specify its green mode, but green_mode kwarg is provided,
        # then we use green_mode kwarg
        (Device1GreenModeUnspecified, GreenMode.Synchronous, GreenMode.Asyncio, None, SynchronousExecutor),
        (Device1GreenModeUnspecified, GreenMode.Synchronous, GreenMode.Gevent, None, SynchronousExecutor),
        (Device1GreenModeUnspecified, GreenMode.Asyncio, GreenMode.Synchronous, None, AsyncioExecutor),
        (Device1GreenModeUnspecified, GreenMode.Asyncio, GreenMode.Gevent, None, AsyncioExecutor),
        (Device1GreenModeUnspecified, GreenMode.Gevent, GreenMode.Synchronous, None, GeventExecutor),
        (Device1GreenModeUnspecified, GreenMode.Gevent, GreenMode.Asyncio, None, GeventExecutor),

        # Finally, if neither device green mode nor green_mode kwarg are specified, then use global mode instead.
        # (currently only works for synchronous mode - see unsupported modes below)
        (Device1GreenModeUnspecified, None, GreenMode.Synchronous, None, SynchronousExecutor),

        # Unsupported modes - device servers with the following combinations
        # fail to start up. The cause is unknown. This could be fixed in the future.
        (Device1GreenModeUnspecified, None, GreenMode.Asyncio, RuntimeError, AsyncioExecutor),
        (Device1GreenModeUnspecified, None, GreenMode.Gevent, RuntimeError, GeventExecutor),
        (Device1Asyncio, None,  GreenMode.Asyncio, RuntimeError, AsyncioExecutor),
        (Device1Gevent, None, GreenMode.Gevent, RuntimeError, GeventExecutor),
    ]
)
def test_green_modes_in_device_kwarg_and_global(
    device_type, green_mode, global_mode, exception_type, executor_type
):
    if WINDOWS and exception_type is not None:
        pytest.skip("Skip test that hangs on Windows")
        return

    old_green_mode = tango.get_green_mode()
    try:
        tango.set_green_mode(global_mode)

        if exception_type is None:
            with DeviceTestContext(device_type, green_mode=green_mode):
                pass
            assert type(get_worker()) == executor_type
        else:
            with pytest.raises(exception_type, match=r"stuck at init"):
                with DeviceTestContext(device_type, green_mode=green_mode, timeout=0.5):
                    pass

    finally:
        tango.set_green_mode(old_green_mode)


def test_multi_with_async_devices_initialised():

    class TestDevice2Async(Device2):
        green_mode = GreenMode.Asyncio

    devices_info = (
        {"class": Device1AsyncInit, "devices": [{"name": "test/device1/1"}]},
        {"class": TestDevice2Async, "devices": [{"name": "test/device2/1"}]},
    )

    with MultiDeviceTestContext(devices_info) as context:
        proxy1 = context.get_device("test/device1/1")
        proxy2 = context.get_device("test/device2/1")
        assert proxy1.State() == tango.DevState.ON
        assert proxy2.State() == tango.DevState.ON
        assert proxy1.attr1 == 150
        assert proxy2.attr2 == 200


def test_multi_device_access():
    devices_info = (
        {"class": Device1, "devices": [{"name": "test/device1/1"}]},
        {"class": Device2, "devices": [{"name": "test/device2/2"}]},
    )

    with MultiDeviceTestContext(devices_info) as context:
        device_access1 = context.get_device_access("test/device1/1")
        device_access2 = context.get_device_access("test/device2/2")
        server_access = context.get_server_access()
        assert "test/device1/1" in device_access1
        assert "test/device2/2" in device_access2
        assert context.server_name in server_access
        proxy1 = tango.DeviceProxy(device_access1)
        proxy2 = tango.DeviceProxy(device_access2)
        proxy_server = tango.DeviceProxy(server_access)
        assert proxy1.attr1 == 100
        assert proxy2.attr2 == 200
        assert proxy_server.State() == tango.DevState.ON


def test_multi_device_proxy_cached():
    devices_info = ({"class": Device1, "devices": [{"name": "test/device1/1"}]},)

    with MultiDeviceTestContext(devices_info) as context:
        device1_first = context.get_device("test/device1/1")
        device1_second = context.get_device("test/device1/1")
        assert device1_first is device1_second


def test_multi_with_two_devices_with_properties(server_green_mode):
    class TestDevice1(Device):
        green_mode = server_green_mode

        prop1 = device_property(dtype=str)

        @command(dtype_out=str)
        def get_prop1(self):
            return self.prop1

    class TestDevice2(Device):
        green_mode = server_green_mode

        prop2 = device_property(dtype=int)

        @command(dtype_out=int)
        def get_prop2(self):
            return self.prop2

    devices_info = (
        {
            "class": TestDevice1,
            "devices": [{"name": "test/device1/1", "properties": {"prop1": "abcd"}}],
        },
        {
            "class": TestDevice2,
            "devices": [{"name": "test/device2/2", "properties": {"prop2": 5555}}],
        },
    )

    with MultiDeviceTestContext(devices_info) as context:
        proxy1 = context.get_device("test/device1/1")
        proxy2 = context.get_device("test/device2/2")
        assert proxy1.get_prop1() == "abcd"
        assert proxy2.get_prop2() == 5555


@pytest.fixture(
    # Per test we have the input config tuple, and then the expected exception type
    params=[
        # empty config
        [tuple(), IndexError],
        # missing/invalid keys
        [({"not-class": Device1, "devices": [{"name": "test/device1/1"}]},), KeyError],
        [({"class": Device1, "not-devices": [{"name": "test/device1/1"}]},), KeyError],
        [({"class": Device1, "devices": [{"not-name": "test/device1/1"}]},), KeyError],
        # duplicate class
        [
            (
                {"class": Device1, "devices": [{"name": "test/device1/1"}]},
                {"class": Device1, "devices": [{"name": "test/device1/2"}]},
            ),
            ValueError,
        ],
        # mixing old "classic" API and new high level API
        [
            (
                {"class": Device1, "devices": [{"name": "test/device1/1"}]},
                {
                    "class": (ClassicAPISimpleDeviceClass, ClassicAPISimpleDeviceImpl),
                    "devices": [{"name": "test/device1/2"}],
                },
            ),
            ValueError,
        ],
        # mixing green modes
        [
            (
                {"class": Device1Synchronous, "devices": [{"name": "test/device1/1"}]},
                {"class": Device1Gevent, "devices": [{"name": "test/device1/2"}]},
            ),
            ValueError,
        ],
    ]
)
def bad_multi_device_config(request):
    return request.param


def test_multi_bad_config_fails(bad_multi_device_config):
    bad_config, expected_error = bad_multi_device_config
    with pytest.raises(expected_error):
        with MultiDeviceTestContext(bad_config):
            pass


@pytest.fixture()
def memorized_attribute_test_device_factory():
    """
    Returns a test device factory that provides a test device with an
    attribute that is memorized or not, according to its boolean
    argument
    """
    def _factory(is_attribute_memorized):
        class _Device(Device):
            def init_device(self):
                self._attr_value = 0

            attr = attribute(
                access=AttrWriteType.READ_WRITE,
                memorized=is_attribute_memorized,
                hw_memorized=is_attribute_memorized
            )

            def read_attr(self):
                return self._attr_value

            def write_attr(self, value):
                self._attr_value = value

        return _Device
    return _factory


@pytest.mark.parametrize(
    "is_attribute_memorized, memorized_value, expected_value",
    [
        (False, None, 0),
        (False, "1", 0),
        (True, None, 0),
        (True, "1", 1),
    ]
)
def test_multi_with_memorized_attribute_values(
    memorized_attribute_test_device_factory,
    is_attribute_memorized,
    memorized_value,
    expected_value
):
    TestDevice = memorized_attribute_test_device_factory(is_attribute_memorized)

    device_info = {"name": "test/device1/1"}
    if memorized_value is not None:
        device_info["memorized"] = {"attr": memorized_value}

    devices_info = (
        {
            "class": TestDevice,
            "devices": [device_info]
        },
    )

    with MultiDeviceTestContext(devices_info) as context:
        proxy = context.get_device("test/device1/1")
        assert proxy.attr == expected_value


@pytest.mark.parametrize(
    "is_attribute_memorized, memorized_value, expected_value",
    [
        (False, None, 0),
        (False, 1, 0),
        (True, None, 0),
        (True, 1, 1),
    ]
)
def test_single_with_memorized_attribute_values(
    memorized_attribute_test_device_factory,
    is_attribute_memorized,
    memorized_value,
    expected_value
):
    TestDevice = memorized_attribute_test_device_factory(is_attribute_memorized)

    kwargs = {
        "memorized": {"attr": memorized_value}
    } if memorized_value is not None else {}

    with DeviceTestContext(TestDevice, **kwargs) as proxy:
        assert proxy.attr == expected_value



#!/usr/bin/env python3

"""
Simple power supply "hardware" simulator, communicating with plain
text commands over TCP.  For uses with the ps1*.py and ps2*.py device
servers.

Requires gevent to be installed.
"""

import time
import logging
import gevent.server

DEFAULT_BIND = ''
DEFAULT_PORT = 45000


class Attr:
    def __init__(self, *, initial_value=0.,
                 encode=lambda x: bytes(str(x), 'ascii'),
                 decode=float):
        self.value = initial_value
        self.encode = encode
        self.decode = decode

    def get(self):
        return self.encode(self.value)

    def set(self, value):
        self.value = self.decode(value)


class Calibrate(Attr):

    def set(self, value):
        self.ts = time.time()
        super().set(value)


class State(Attr):
    def __init__(self, calib, *args, **kwargs):
        kwargs['initial_value'] = 0
        kwargs['decode'] = int
        super().__init__(*args, **kwargs)
        self.calib = calib
        calib.ts = 0

    def get(self):
        self.value = 0
        if time.time() - self.calib.ts < 2:
            self.value = 1
        return super().get()


class PSSimulator(gevent.server.StreamServer):

    class Error(Exception):
        pass

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log = logging.getLogger(f'simulator.{self.server_port}')
        calib = Calibrate(initial_value=0)
        self.attrs = {
            b'stat': State(calib),
            b'vol': Attr(initial_value=0.1),
            b'curr': Attr(initial_value=0.),
            b'calib': calib,
        }

    def __getitem__(self, name):
        return self.attrs[name].get()

    def __setitem__(self, name, value):
        self.attrs[name].set(value)

    def handle(self, sock, addr):
        log = self.log
        log.info('new connection from %r', addr)
        fileobj = sock.makefile(mode='rb')
        while True:
            request = fileobj.readline()
            if not request:
                log.info('disconnected %r', addr)
                break
            log.info('request %r', request)
            reply = b'ERROR'
            try:
                reply = self.handle_request(request)
            except PSSimulator.Error:
                pass
            except Exception:
                log.exception('Unforeseen error')
            gevent.sleep(1e-1)
            sock.sendall(reply + b'\n')
            log.info('replyed %r', reply)
        fileobj.close()

    def handle_request(self, request):
        req_lower = request.strip().lower()
        is_query = b'?' in req_lower
        pars = req_lower.split()
        name = pars[0]
        if is_query:
            name = name[:-1] # take out '?'
        if is_query:
            return self[name]
        self[name] = pars[1]
        return b'OK'


def main(number=1, bind=DEFAULT_BIND, port=DEFAULT_PORT, **kwargs):
    servers = []
    logging.info('starting simulator...')
    for i in range(number):
        address = bind, port+i
        server = PSSimulator(address)
        server.start()
        servers.append(server)
        server.log.info('simulator listenning on %r!', address)
    try:
        while True:
            gevent.sleep(1)
    except KeyboardInterrupt:
        logging.info('Ctrl-C pressed. Bailing out!')
        for server in servers:
            server.stop()


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--port', type=int, default=DEFAULT_PORT)
    parser.add_argument('--bind', default=DEFAULT_BIND)
    parser.add_argument('--log-level', default='info')
    parser.add_argument('--number', type=int, default=1)
    args = parser.parse_args()
    logging.basicConfig(level=args.log_level.upper())
    main(**vars(args))



# ------------------------------------------------------------------------------
# This file is part of PyTango (http://pytango.rtfd.io)
#
# Copyright 2006-2012 CELLS / ALBA Synchrotron, Bellaterra, Spain
# Copyright 2013-2014 European Synchrotron Radiation Facility, Grenoble, France
#
# Distributed under the terms of the GNU Lesser General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
# ------------------------------------------------------------------------------

"""
This is an internal PyTango module.
"""

__all__ = ("auto_monitor_init",)

__docformat__ = "restructuredtext"

from ._tango import AutoTangoMonitor, AutoTangoAllowThreads


def __AutoTangoMonitor__enter__(self):
    self._acquire()
    return self


def __AutoTangoMonitor__exit__(self, *args, **kwargs):
    self._release()


def __init_AutoTangoMonitor():
    AutoTangoMonitor.__enter__ = __AutoTangoMonitor__enter__
    AutoTangoMonitor.__exit__ = __AutoTangoMonitor__exit__


def __doc_AutoTangoMonitor():
    AutoTangoMonitor.__doc__ = """\

    In a tango server, guard the tango monitor within a python context::

        with AutoTangoMonitor(dev):
            # code here is protected by the tango device monitor
            do something
    """


def __AutoTangoAllowThreads__enter__(self):
    return self


def __AutoTangoAllowThreads__exit__(self, *args, **kwargs):
    self._acquire()


def __init_AutoTangoAllowThreads():
    AutoTangoAllowThreads.__enter__ = __AutoTangoAllowThreads__enter__
    AutoTangoAllowThreads.__exit__ = __AutoTangoAllowThreads__exit__


def __doc_AutoTangoAllowThreads():
    AutoTangoAllowThreads.__doc__ = """\

    In a tango server, free the tango monitor within a context:

        with AutoTangoAllowThreads(dev):
            # code here is not under the tango device monitor
            do something
    """


def auto_monitor_init(doc=True):
    __init_AutoTangoMonitor()
    __init_AutoTangoAllowThreads()
    if doc:
        __doc_AutoTangoMonitor()
        __doc_AutoTangoAllowThreads()



import tango
from tango.server import Device
from tango.server import pipe
from tango import DevState


class PipeServer(Device):

    def init_device(self):
        Device.init_device(self)
        self.rootBlobName = 'theBlob'
        self.__blob = self.rootBlobName, dict(x=0.3, y=10.22,
                                              width=105.1, height=206.6)

    @pipe(label="Test pipe", fisallowed="is_TestPipe_allowed")
    def TestPipe(self):
        return self.__blob

    @TestPipe.write
    def TestPipe(self, blob):
        self.__blob = blob
        print(blob)

    def is_TestPipe_allowed(self, pipeReqType):
        """ pipeReqType is either READ_REQ or WRITE_REQ."""
        if pipeReqType == tango.AttReqType.READ_REQ:
            return self.get_state() not in [DevState.FAULT,
                                            DevState.OFF]
        else:
            return self.get_state() not in [DevState.FAULT,
                                            DevState.OFF,
                                            DevState.MOVING]

# ----------
# Run server
# ----------


if __name__ == '__main__':
    PipeServer.run_server()



#!/usr/bin/env python
# ------------------------------------------------------------------------------
# This file is part of PyTango (http://pytango.rtfd.io)
#
# Copyright 2006-2012 CELLS / ALBA Synchrotron, Bellaterra, Spain
# Copyright 2013-2014 European Synchrotron Radiation Facility, Grenoble, France
#
# Distributed under the terms of the GNU Lesser General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# WARNING: This script should only be executed as a Post-Build Event from inside
#          Microsoft Visual Studio and not from the command line
# ------------------------------------------------------------------------------


import sys
import os
import os.path as osp


def main():
    executable = sys.executable

    curr_dir = os.getcwd()

    winsetup_dir = osp.dirname(osp.abspath(__file__))
    os.chdir(winsetup_dir)
    setup_name = "setup.py"
    bitmap = osp.join(winsetup_dir, 'doc', 'logo-medium.bmp')
    ver = ".".join(map(str, sys.version_info[:2]))

    print("winsetup: invoked with: " + ' '.join(sys.argv))
    if len(sys.argv) < 6:
        print("winsetup: need to supply build directory, distribution directory, temporary binary install directory, configuration name and platform name")
        return 1

    build_dir, dist_dir, bdist_dir = map(osp.abspath, sys.argv[1:4])
    config_name, plat_name = sys.argv[4:6]
    # Pypi is picky about platform name. Make sure we obey his/her majesty
    plat_name = plat_name.lower()
#    temp_base_dir = osp.abspath(os.environ["TEMP"])
#    temp_dir = osp.join(temp_base_dir, "PyTango", config_name)
    if plat_name == 'x64':
        plat_name = 'win-amd64'

    try:
        cmd_line = f'{executable} {setup_name} '
        cmd_line += f'build_py --force --no-compile --build-lib={build_dir} '
        cmd_line += 'build_scripts --force '
        cmd_line += f'install_lib --skip-build --no-compile --build-dir={build_dir} '
#        cmd_line += 'bdist_msi --skip-build --target-version=%s ' \
#                    '--bdist-dir=%s ' \
#                    '--dist-dir=%s ' \
#                    '--plat-name=%s ' % (ver, bdist_dir, dist_dir, plat_name)
        cmd_line += f'bdist_wininst --skip-build --target-version={ver} ' \
                    f'--bdist-dir={bdist_dir} ' \
                    f'--dist-dir={dist_dir} ' \
                    f'--title="PyTango 9" ' \
                    f'--bitmap="{bitmap}" ' \
                    f'--plat-name={plat_name} '
        cmd_line += f'bdist_wheel --skip-build ' \
                    f'--bdist-dir={bdist_dir} ' \
                    f'--dist-dir={dist_dir} ' \
                    f'--plat-name={plat_name} '
        os.system(cmd_line)
    except Exception:
        print("Failed:")
        import traceback
        traceback.print_exc()
        return 2
    finally:
        os.chdir(curr_dir)

    return 0


if __name__ == "__main__":
    ret = main()
    sys.exit(ret)



# ------------------------------------------------------------------------------
# This file is part of PyTango (http://pytango.rtfd.io)
#
# Copyright 2006-2012 CELLS / ALBA Synchrotron, Bellaterra, Spain
# Copyright 2013-2014 European Synchrotron Radiation Facility, Grenoble, France
#
# Distributed under the terms of the GNU Lesser General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
# ------------------------------------------------------------------------------

"""
This is an internal PyTango module. It completes the binding of
:class:`tango.AttributeProxy`.

To access these members use directly :mod:`tango` module and NOT
tango.attribute_proxy.
"""

import collections.abc

from ._tango import StdStringVector, DbData, DbDatum, DeviceProxy
from ._tango import __AttributeProxy as _AttributeProxy
from .utils import seq_2_StdStringVector, seq_2_DbData, DbData_2_dict
from .utils import is_pure_str, is_non_str_seq
from .green import green, get_green_mode
from .device_proxy import __init_device_proxy_internals as init_device_proxy

__all__ = ("AttributeProxy", "attribute_proxy_init", "get_attribute_proxy")


@green(consume_green_mode=False)
def get_attribute_proxy(*args, **kwargs):
    """
    get_attribute_proxy(self, full_attr_name, green_mode=None, wait=True, timeout=True) -> AttributeProxy
    get_attribute_proxy(self, device_proxy, attr_name, green_mode=None, wait=True, timeout=True) -> AttributeProxy

    Returns a new :class:`~tango.AttributeProxy`.
    There is no difference between using this function and the direct
    :class:`~tango.AttributeProxy` constructor if you use the default kwargs.

    The added value of this function becomes evident when you choose a green_mode
    to be *Futures* or *Gevent*. The AttributeProxy constructor internally makes some
    network calls which makes it *slow*. By using one of the *green modes* as
    green_mode you are allowing other python code to be executed in a cooperative way.

    :param full_attr_name: the full name of the attribute
    :type full_attr_name: str
    :param device_proxy: the :class:`~tango.DeviceProxy`
    :type device_proxy: DeviceProxy
    :param attr_name: attribute name for the given device proxy
    :type attr_name: str
    :param green_mode: determines the mode of execution of the device (including
                      the way it is created). Defaults to the current global
                      green_mode (check :func:`~tango.get_green_mode` and
                      :func:`~tango.set_green_mode`)
    :type green_mode: :obj:`~tango.GreenMode`
    :param wait: whether or not to wait for result. If green_mode
                 Ignored when green_mode is Synchronous (always waits).
    :type wait: bool
    :param timeout: The number of seconds to wait for the result.
                    If None, then there is no limit on the wait time.
                    Ignored when green_mode is Synchronous or wait is False.
    :type timeout: float
    :returns:
        if green_mode is Synchronous or wait is True:
            :class:`~tango.AttributeProxy`
        else if green_mode is Futures:
            :class:`concurrent.futures.Future`
        else if green_mode is Gevent:
            :class:`gevent.event.AsynchResult`
    :throws:
        * a *DevFailed* if green_mode is Synchronous or wait is True
          and there is an error creating the attribute.
        * a *concurrent.futures.TimeoutError* if green_mode is Futures,
          wait is False, timeout is not None and the time to create the attribute
          has expired.
        * a *gevent.timeout.Timeout* if green_mode is Gevent, wait is False,
          timeout is not None and the time to create the attribute has expired.

    New in PyTango 8.1.0
    """
    return AttributeProxy(*args, **kwargs)


def __AttributeProxy__get_property(self, propname, value=None):
    """
    get_property(self, propname, value) -> DbData

            Get a (list) property(ies) for an attribute.

            This method accepts the following types as propname parameter:
            1. string [in] - single property data to be fetched
            2. sequence<string> [in] - several property data to be fetched
            3. tango.DbDatum [in] - single property data to be fetched
            4. tango.DbData [in,out] - several property data to be fetched.
            5. sequence<DbDatum> - several property data to be feteched

            Note: for cases 3, 4 and 5 the 'value' parameter if given, is IGNORED.

            If value is given it must be a tango.DbData that will be filled with the
            property values

        Parameters :
            - propname : (str) property(ies) name(s)
            - value : (tango.DbData) (optional, default is None meaning that the
                      method will create internally a tango.DbData and return
                      it filled with the property values

        Return     : (DbData) containing the property(ies) value(s). If a
                     tango.DbData is given as parameter, it returns the same
                     object otherwise a new tango.DbData is returned

        Throws     : NonDbDevice, ConnectionFailed (with database),
                     CommunicationFailed (with database),
                     DevFailed from database device
    """

    if is_pure_str(propname) or isinstance(propname, StdStringVector):
        new_value = value
        if new_value is None:
            new_value = DbData()
        self._get_property(propname, new_value)
        return DbData_2_dict(new_value)
    elif isinstance(propname, DbDatum):
        new_value = DbData()
        new_value.append(propname)
        self._get_property(new_value)
        return DbData_2_dict(new_value)
    elif isinstance(propname, collections.abc.Sequence):
        if isinstance(propname, DbData):
            self._get_property(propname)
            return DbData_2_dict(propname)

        if is_pure_str(propname[0]):
            new_propname = StdStringVector()
            for i in propname:
                new_propname.append(i)
            new_value = value
            if new_value is None:
                new_value = DbData()
            self._get_property(new_propname, new_value)
            return DbData_2_dict(new_value)
        elif isinstance(propname[0], DbDatum):
            new_value = DbData()
            for i in propname:
                new_value.append(i)
            self._get_property(new_value)
            return DbData_2_dict(new_value)


def __AttributeProxy__put_property(self, value):
    """
    put_property(self, value) -> None

            Insert or update a list of properties for this attribute.
            This method accepts the following types as value parameter:
            1. tango.DbDatum - single property data to be inserted
            2. tango.DbData - several property data to be inserted
            3. sequence<DbDatum> - several property data to be inserted
            4. dict<str, DbDatum> - keys are property names and value has data to be inserted
            5. dict<str, seq<str>> - keys are property names and value has data to be inserted
            6. dict<str, obj> - keys are property names and str(obj) is property value

        Parameters :
            - value : can be one of the following:
                1. tango.DbDatum - single property data to be inserted
                2. tango.DbData - several property data to be inserted
                3. sequence<DbDatum> - several property data to be inserted
                4. dict<str, DbDatum> - keys are property names and value has data to be inserted
                5. dict<str, seq<str>> - keys are property names and value has data to be inserted
                6. dict<str, obj> - keys are property names and str(obj) is property value

        Return     : None

        Throws     : ConnectionFailed, CommunicationFailed
                     DevFailed from device (DB_SQLError),
                     TypeError
    """
    if isinstance(value, DbData):
        pass
    elif isinstance(value, DbDatum):
        new_value = DbData()
        new_value.append(value)
        value = new_value
    elif is_non_str_seq(value):
        new_value = seq_2_DbData(value)
    elif isinstance(value, collections.abc.Mapping):
        new_value = DbData()
        for k, v in value.items():
            if isinstance(v, DbDatum):
                new_value.append(v)
                continue
            db_datum = DbDatum(k)
            if is_non_str_seq(v):
                seq_2_StdStringVector(v, db_datum.value_string)
            else:
                db_datum.value_string.append(str(v))
            new_value.append(db_datum)
        value = new_value
    else:
        raise TypeError(
            'Value must be a tango.DbDatum, tango.DbData, '
            'a sequence<DbDatum> or a dictionary')
    return self._put_property(value)


def __AttributeProxy__delete_property(self, value):
    """
    delete_property(self, value) -> None

        Delete a the given of properties for this attribute.
        This method accepts the following types as value parameter:

            1. string [in] - single property to be deleted
            2. tango.DbDatum [in] - single property data to be deleted
            3. tango.DbData [in] - several property data to be deleted
            4. sequence<string> [in]- several property data to be deleted
            5. sequence<DbDatum> [in] - several property data to be deleted
            6. dict<str, obj> [in] - keys are property names to be deleted
               (values are ignored)
            7. dict<str, DbDatum> [in] - several DbDatum.name are property names
               to be deleted (keys are ignored)

        Parameters :
            - value : can be one of the following:

                1. string [in] - single property data to be deleted
                2. tango.DbDatum [in] - single property data to be deleted
                3. tango.DbData [in] - several property data to be deleted
                4. sequence<string> [in]- several property data to be deleted
                5. sequence<DbDatum> [in] - several property data to be deleted
                6. dict<str, obj> [in] - keys are property names to be deleted
                   (values are ignored)
                7. dict<str, DbDatum> [in] - several DbDatum.name are property
                   names to be deleted (keys are ignored)

        Return     : None

        Throws     : ConnectionFailed, CommunicationFailed
                     DevFailed from device (DB_SQLError),
                     TypeError
    """
    if isinstance(value, DbData) or isinstance(value, StdStringVector) or \
            is_pure_str(value):
        new_value = value
    elif isinstance(value, DbDatum):
        new_value = DbData()
        new_value.append(value)
    elif isinstance(value, collections.abc.Sequence):
        new_value = DbData()
        for e in value:
            if isinstance(e, DbDatum):
                new_value.append(e)
            else:
                new_value.append(DbDatum(str(e)))
    elif isinstance(value, collections.abc.Mapping):
        new_value = DbData()
        for k, v in value.items():
            if isinstance(v, DbDatum):
                new_value.append(v)
            else:
                new_value.append(DbDatum(k))
    else:
        raise TypeError(
            'Value must be a string, tango.DbDatum, '
            'tango.DbData, a sequence or a dictionary')

    return self._delete_property(new_value)


# It is easier to reimplement AttributeProxy in python using DeviceProxy than
# wrapping C++ AttributeProxy. However I still rely in the original
# AttributeProxy for the constructor (parsing strings if necessary) and some
# other things. With the _method_* functions defined later it is really easy.
# One reason to do it this way: get_device_proxy() will always return the
# same tango.DeviceProxy with this implementation. And then we can trust
# it's automatic event unsubscription to handle events.
class AttributeProxy:
    """
        AttributeProxy is the high level Tango object which provides the
        client with an easy-to-use interface to TANGO attributes.

        To create an AttributeProxy, a complete attribute name must be set
        in the object constructor.

        Example:
            att = AttributeProxy("tango/tangotest/1/long_scalar")

        Note: PyTango implementation of AttributeProxy is in part a
        python reimplementation of the AttributeProxy found on the C++ API.
    """

    def __init__(self, *args, **kwds):
        green_mode = kwds.pop('green_mode', get_green_mode())
        self.__attr_proxy = _AttributeProxy(*args, **kwds)
        # get_device_proxy() returns a different python object each time
        # we don't want a different object, so we save the current one.
        self.__dev_proxy = dp = self.__attr_proxy.get_device_proxy()
        init_device_proxy(dp)
        dp.set_green_mode(green_mode)

    def get_device_proxy(self):
        """
        get_device_proxy(self) -> DeviceProxy

                A method which returns the device associated to the attribute

            Parameters : None

            Return     : (DeviceProxy)
        """
        return self.__dev_proxy

    def name(self):
        """
        name(self) -> str

                Returns the attribute name

            Parameters : None
            Return     : (str) with the attribute name
        """
        return self.__attr_proxy.name()

    def __str__(self):
        return f"AttributeProxy({self.name()})"

    def __repr__(self):
        return f"AttributeProxy({self.name()})"


def _method_dev_and_name(dp_fn_name, doc=True):
    def __new_fn(self, *args, **kwds):
        return getattr(self._AttributeProxy__dev_proxy, dp_fn_name)(self.name(), *args, **kwds)

    if doc:
        __new_fn.__doc__ = "This method is a simple way to do:\n" + \
                           "\tself.get_device_proxy()." + dp_fn_name + \
                           "(self.name(), ...)\n\n" + \
                           "For convenience, here is the documentation of DeviceProxy." + \
                           dp_fn_name + "(...):\n" + \
                           str(getattr(DeviceProxy, dp_fn_name).__doc__)
    __new_fn.__name__ = dp_fn_name
    return __new_fn


def _method_device(dp_fn_name, doc=True):
    def __new_fn(self, *args, **kwds):
        return getattr(self._AttributeProxy__dev_proxy, dp_fn_name)(*args, **kwds)

    if doc:
        __new_fn.__doc__ = "This method is a simple way to do:\n" + \
                           "\tself.get_device_proxy()." + dp_fn_name + \
                           "(...)\n\n" + \
                           "For convenience, here is the documentation of DeviceProxy." + \
                           dp_fn_name + "(...):\n" + \
                           str(getattr(DeviceProxy, dp_fn_name).__doc__)
    __new_fn.__name__ = dp_fn_name
    return __new_fn


def _method_attribute(dp_fn_name, doc=True):
    def __new_fn(self, *args, **kwds):
        return getattr(self._AttributeProxy__attr_proxy, dp_fn_name)(*args, **kwds)

    if doc:
        __new_fn.__doc__ = getattr(_AttributeProxy, dp_fn_name).__doc__
    __new_fn.__name__ = dp_fn_name
    return __new_fn


def __init_AttributeProxy(doc=True):
    _AttributeProxy.get_property = __AttributeProxy__get_property
    _AttributeProxy.put_property = __AttributeProxy__put_property
    _AttributeProxy.delete_property = __AttributeProxy__delete_property

    # General methods
    # AttributeProxy.name                manually defined
    AttributeProxy.status = _method_device('status', doc=doc)
    AttributeProxy.state = _method_device('state', doc=doc)
    AttributeProxy.ping = _method_device('ping', doc=doc)
    AttributeProxy.get_transparency_reconnection = _method_device('get_transparency_reconnection', doc=doc)
    AttributeProxy.set_transparency_reconnection = _method_device('set_transparency_reconnection', doc=doc)

    # Property methods
    AttributeProxy.get_property = _method_attribute('get_property', doc=doc)
    AttributeProxy.put_property = _method_attribute('put_property', doc=doc)
    AttributeProxy.delete_property = _method_attribute('delete_property', doc=doc)

    # Attribute methods
    AttributeProxy.get_config = _method_dev_and_name('get_attribute_config', doc=doc)
    AttributeProxy.set_config = _method_device('set_attribute_config', doc=doc)

    AttributeProxy.write = _method_dev_and_name('write_attribute', doc=doc)
    AttributeProxy.read = _method_dev_and_name('read_attribute', doc=doc)
    AttributeProxy.write_read = _method_dev_and_name('write_read_attribute', doc=doc)

    # History methods...
    AttributeProxy.history = _method_dev_and_name('attribute_history', doc=doc)

    # Polling administration methods
    AttributeProxy.poll = _method_dev_and_name('poll_attribute', doc=doc)
    AttributeProxy.get_poll_period = _method_dev_and_name('get_attribute_poll_period', doc=doc)
    AttributeProxy.is_polled = _method_dev_and_name('is_attribute_polled', doc=doc)
    AttributeProxy.stop_poll = _method_dev_and_name('stop_poll_attribute', doc=doc)

    # Asynchronous methods
    AttributeProxy.read_asynch = _method_dev_and_name('read_attribute_asynch', doc=doc)
    AttributeProxy.read_reply = _method_device('read_attribute_reply', doc=doc)
    AttributeProxy.write_asynch = _method_device('write_attribute_asynch', doc=doc)
    AttributeProxy.write_reply = _method_device('write_attribute_reply', doc=doc)

    # Event methods
    AttributeProxy.subscribe_event = _method_dev_and_name('subscribe_event', doc=doc)
    AttributeProxy.unsubscribe_event = _method_device('unsubscribe_event', doc=doc)

    AttributeProxy.get_events = _method_device('get_events', doc=doc)
    AttributeProxy.event_queue_size = _method_device('event_queue_size', doc=doc)
    AttributeProxy.get_last_event_date = _method_device('get_last_event_date', doc=doc)
    AttributeProxy.is_event_queue_empty = _method_device('is_event_queue_empty', doc=doc)


def attribute_proxy_init(doc=True):
    __init_AttributeProxy(doc=doc)



"""Test utilities"""

import enum
import numpy as np

# Local imports
from . import DevState, GreenMode, AttrDataFormat, ExtractAs
from .server import Device
from .test_context import (
    MultiDeviceTestContext, DeviceTestContext, get_server_port_via_pid
)
from .utils import is_non_str_seq, FROM_TANGO_TO_NUMPY_TYPE
from . import DeviceClass, LatestDeviceImpl, DevLong64, SCALAR, READ

# Conditional imports
try:
    import pytest
except ImportError:
    pytest = None

__all__ = (
    'MultiDeviceTestContext',
    'DeviceTestContext',
    'SimpleDevice',
    'ClassicAPISimpleDeviceImpl',
    'ClassicAPISimpleDeviceClass',
    'state',
    'command_typed_values',
    'attribute_typed_values',
    'server_green_mode',
    'attr_data_format',
    'get_server_port_via_pid',
)

# char \x00 cannot be sent in a DevString. All other 1-255 chars can
ints = tuple(range(1, 256))
bytes_devstring = bytes(ints)
str_devstring = bytes_devstring.decode('latin-1')

# Test devices

class SimpleDevice(Device):
    def init_device(self):
        self.set_state(DevState.ON)


class ClassicAPISimpleDeviceImpl(LatestDeviceImpl):
    def __init__(self, cls, name):
        LatestDeviceImpl.__init__(self, cls, name)
        ClassicAPISimpleDeviceImpl.init_device(self)

    def init_device(self):
        self.get_device_properties(self.get_device_class())
        self.attr_attr1_read = 100

    def read_attr1(self, attr):
        attr.set_value(self.attr_attr1_read)


class ClassicAPISimpleDeviceClass(DeviceClass):
    attr_list = {"attr1": [[DevLong64, SCALAR, READ]]}


# Test enums

class GoodEnum(enum.IntEnum):
    START = 0
    MIDDLE = 1
    END = 2


class BadEnumNonZero(enum.IntEnum):
    START = 1
    MIDDLE = 2
    END = 3


class BadEnumSkipValues(enum.IntEnum):
    START = 0
    MIDDLE = 2
    END = 4


class BadEnumDuplicates(enum.IntEnum):
    START = 0
    MIDDLE = 1
    END = 1


# Helpers

# Note on Tango properties using the Tango File database:
# Tango file database cannot handle properties with '\n'. It doesn't
# handle '\' neither. And it cuts ASCII extended characters. That is
# why you will find that all property related tests are truncated to
# the first two values of the arrays below

GENERAL_TYPED_VALUES = {
    int: (1, 2, -65535, 23),
    float: (2.71, 3.14, -34.678e-10, 12.678e+15),
    str: ('hey hey', 'my my', bytes_devstring, str_devstring),
    bool: (False, True, True, False),
    (int,): (np.array([1, 2]), [1, 2, 3], [9, 8, 7], [-65535, 2224], [0, 0], ),
    (float,): (np.array([0.1, 0.2]), [0.1, 0.2, 0.3], [0.9, 0.8, 0.7], [-6.3232e-3], [0.0, 12.56e+12]),
    (str,): (np.array(['foo', 'bar']), ['ab', 'cd', 'ef'], ['gh', 'ij', 'kl'], 10*[bytes_devstring], 10*[str_devstring]),
    (bool,): (np.array([True, False]), [False, False, True], [True, False, False], [False], [True]),
}

IMAGE_TYPED_VALUES = {
    ((int,),): (np.vstack((np.array([1, 2]), np.array([3, 4]))),
                [[1, 2, 3], [4, 5, 6]], [[-65535, 2224], [-65535, 2224]],),
    ((float,),): (np.vstack((np.array([0.1, 0.2]), np.array([0.3, 0.4]))),
                  [[0.1, 0.2, 0.3], [0.9, 0.8, 0.7]], [[-6.3232e-3, 0.0], [0.0, 12.56e+12]],),
    ((str,),): (np.vstack((np.array(['hi-hi', 'ha-ha']), np.array(['hu-hu', 'yuhuu']))),
                [['ab', 'cd', 'ef'], ['gh', 'ij', 'kl']], [10*[bytes_devstring],10*[bytes_devstring]],
                [10*[str_devstring], 10*[str_devstring]],),
    ((bool,),): (np.vstack((np.array([True, False]), np.array([False, True]))),
                 [[False, False, True], [True, False, False]], [[False]], [[True]],)
}


EXTRACT_AS = [(ExtractAs.Numpy, np.ndarray),
              (ExtractAs.Tuple, tuple),
              (ExtractAs.List, list),
              (ExtractAs.Bytes, bytes),
              (ExtractAs.ByteArray, bytearray),
              (ExtractAs.String, str)
              ]

BASE_TYPES = [float, int, str, bool]

# these sets to test Device Server input arguments

OS_SYSTEMS = ['linux', 'win']

#    os_system, in string, out arguments list, raised exception
DEVICE_SERVER_ARGUMENTS = (
    (['linux', 'win'], 'MyDs instance --nodb --port 1234',
     ['MyDs', 'instance', '-nodb', '-ORBendPoint', 'giop:tcp:0.0.0.0:1234']),
    (['linux', 'win'], 'MyDs -port 1234 -host myhost instance',
     ['MyDs', 'instance', '-ORBendPoint', 'giop:tcp:myhost:1234']),
    (['linux', 'win'], 'MyDs instance --ORBendPoint giop:tcp::1234',
     ['MyDs', 'instance', '-ORBendPoint', 'giop:tcp::1234']),
    (['linux', 'win'], 'MyDs instance -nodb -port 1000 -dlist a/b/c;d/e/f',
     ['MyDs', 'instance', '-ORBendPoint', 'giop:tcp:0.0.0.0:1000', '-nodb', '-dlist', 'a/b/c;d/e/f']),
    (['linux', 'win'], 'MyDs instance -file a/b/c',
     ['MyDs', 'instance', '-file=a/b/c']),
    ([], 'MyDs instance -nodb', []),  # this test should always fail
    ([], 'MyDs instance -dlist a/b/c;d/e/f', []),  # this test should always fail

# the most complicated case: verbose
    (['linux', 'win'], 'MyDs instance -vvvv', ['MyDs', 'instance', '-v4']),
    (['linux', 'win'], 'MyDs instance --verbose --verbose --verbose --verbose', ['MyDs', 'instance', '-v4']),
    (['linux', 'win'], 'MyDs instance -v4', ['MyDs', 'instance', '-v4']),
    (['linux', 'win'], 'MyDs instance -v 4', ['MyDs', 'instance', '-v4']),

# some options can be only in win, in linux should be error
    (['win'], 'MyDs instance -dbg -i -s -u', ['MyDs', 'instance', '-dbg', '-i', '-s', '-u']),

# variable ORB options
    (['linux', 'win'], 'MyDs instance -ORBtest1 test1 --ORBtest2 test2',
     ['MyDs', 'instance', '-ORBtest1', 'test1', '-ORBtest2', 'test2']),
    (['linux', 'win'], 'MyDs ORBinstance -ORBtest myORBparam',
     ['MyDs', 'ORBinstance', '-ORBtest', 'myORBparam']),
    (['linux', 'win'], 'MyDs instance -nodb -ORBendPoint giop:tcp:localhost:1234 -ORBendPointPublish giop:tcp:myhost.local:2345',
     ['MyDs', 'instance', '-nodb', '-ORBendPoint', 'giop:tcp:localhost:1234', '-ORBendPointPublish', 'giop:tcp:myhost.local:2345']),
    ([], 'MyDs instance -ORBtest1 test1 --orbinvalid value', []),  # lowercase "orb" should fail
)

def repr_type(x):
    if isinstance(x, (list, tuple)):
        return f'({repr_type(x[0])},)'
    return f'{x.__name__}'


# Numpy helpers

if pytest:

    def __assert_all_types(a, b):
        if isinstance(a, str):
            assert a == b
            return

        try:
            assert a == pytest.approx(b)
        except (ValueError, TypeError):
            np.testing.assert_allclose(a, b)

    def assert_close(a, b):
        if is_non_str_seq(a):
            assert len(a) == len(b)
            for _a, _b in zip(a, b):
                assert_close(_a, _b)
        else:
            __assert_all_types(a, b)


    def __convert_value(value):
        if isinstance(value, bytes):
            return value.decode('latin-1')
        return value

    def create_result(dtype, value):
        if isinstance(dtype, (list, tuple)):
            dtype = dtype[0]
            return [create_result(dtype, v) for v in value]

        return __convert_value(value)

    def convert_to_type(value, attr_type, expected_type):
        if expected_type in [tuple, list]:
            return expected_type(value)
        elif expected_type == np.ndarray:
            return np.array(value, dtype=FROM_TANGO_TO_NUMPY_TYPE[attr_type])
        elif expected_type in [bytes, bytearray, str]:
            value = np.array(value, dtype=FROM_TANGO_TO_NUMPY_TYPE[attr_type]).tobytes()
            if expected_type == bytearray:
                return bytearray(value)
            elif expected_type == str:
                return "".join([chr(b) for b in value])
            return value
        else:
            pytest.xfail("Unknown extract_as type")

    @pytest.fixture(params=DevState.values.values())
    def state(request):
        return request.param

    @pytest.fixture(
        params=list(GENERAL_TYPED_VALUES.items()),
        ids=lambda x: repr_type(x[0]))
    def command_typed_values(request):
        dtype, values = request.param
        expected = lambda v: create_result(dtype, v)
        return dtype, values, expected

    @pytest.fixture(
        params=list({**GENERAL_TYPED_VALUES, **IMAGE_TYPED_VALUES}.items()),
        ids=lambda x: repr_type(x[0]))
    def attribute_typed_values(request):
        dtype, values = request.param
        expected = lambda v: create_result(dtype, v)
        return dtype, values, expected

    @pytest.fixture(params=EXTRACT_AS, ids=[f"extract_as.{req_type}" for req_type, _ in EXTRACT_AS])
    def extract_as(request):
        requested_type, expected_type = request.param
        return requested_type, expected_type

    @pytest.fixture(params=BASE_TYPES)
    def base_type(request):
        return request.param

    @pytest.fixture(params=GreenMode.values.values())
    def green_mode(request):
        return request.param

    @pytest.fixture(params=[
        GreenMode.Synchronous,
        GreenMode.Asyncio,
        GreenMode.Gevent])
    def server_green_mode(request):
        return request.param

    @pytest.fixture(params=[
        AttrDataFormat.SCALAR,
        AttrDataFormat.SPECTRUM,
        AttrDataFormat.IMAGE])
    def attr_data_format(request):
        return request.param



"""Backport some asyncio features."""

import collections
import concurrent.futures
import functools
import inspect
import types

try:
    import asyncio
except ImportError:
    import trollius as asyncio

__all__ = ("run_coroutine_threadsafe",)


def _set_concurrent_future_state(concurrent, source):
    """Copy state from a future to a concurrent.futures.Future."""
    assert source.done()
    if source.cancelled():
        concurrent.cancel()
    if not concurrent.set_running_or_notify_cancel():
        return
    exception = source.exception()
    if exception is not None:
        concurrent.set_exception(exception)
    else:
        result = source.result()
        concurrent.set_result(result)


def _copy_future_state(source, dest):
    """Internal helper to copy state from another Future.
    The other Future may be a concurrent.futures.Future.
    """
    assert source.done()
    if dest.cancelled():
        return
    assert not dest.done()
    if source.cancelled():
        dest.cancel()
    else:
        exception = source.exception()
        if exception is not None:
            dest.set_exception(exception)
        else:
            result = source.result()
            dest.set_result(result)


def _chain_future(source, dest):
    """Chain two futures so that when one completes, so does the other.
    The result (or exception) of source will be copied to destination.
    If destination is cancelled, source gets cancelled too.
    Compatible with both asyncio.Future and concurrent.futures.Future.
    """
    if not isinstance(source, (asyncio.Future, concurrent.futures.Future)):
        raise TypeError('A future is required for source argument')
    if not isinstance(dest, (asyncio.Future, concurrent.futures.Future)):
        raise TypeError('A future is required for destination argument')
    source_loop = source._loop if isinstance(source, asyncio.Future) else None
    dest_loop = dest._loop if isinstance(dest, asyncio.Future) else None

    def _set_state(future, other):
        if isinstance(future, asyncio.Future):
            _copy_future_state(other, future)
        else:
            _set_concurrent_future_state(future, other)

    def _call_check_cancel(destination):
        if destination.cancelled():
            if source_loop is None or source_loop is dest_loop:
                source.cancel()
            else:
                source_loop.call_soon_threadsafe(source.cancel)

    def _call_set_state(source):
        if dest_loop is None or dest_loop is source_loop:
            _set_state(dest, source)
        else:
            dest_loop.call_soon_threadsafe(_set_state, dest, source)

    dest.add_done_callback(_call_check_cancel)
    source.add_done_callback(_call_set_state)


def run_coroutine_threadsafe(coro, loop):
    """Submit a coroutine object to a given event loop.
    Return a concurrent.futures.Future to access the result.
    """
    if not asyncio.iscoroutine(coro):
        raise TypeError('A coroutine object is required')
    future = concurrent.futures.Future()

    def callback():
        try:
            _chain_future(asyncio.ensure_future(coro, loop=loop), future)
        except Exception as exc:
            if future.set_running_or_notify_cancel():
                future.set_exception(exc)
            raise

    loop.call_soon_threadsafe(callback)
    return future


# Function removed from Python 3.11
# Taken from https://github.com/python/cpython/blob/3.10/Lib/asyncio/coroutines.py
# (without the _DEBUG part)
# Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010,
# 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020, 2021, 2022 Python Software Foundation;
# All Rights Reserved
def coroutine(func):
    """Decorator to mark coroutines.
    If the coroutine is not yielded from before it is destroyed,
    an error message is logged.
    """
    if inspect.iscoroutinefunction(func):
        return func

    if inspect.isgeneratorfunction(func):
        coro = func
    else:
        @functools.wraps(func)
        def coro(*args, **kw):
            res = func(*args, **kw)
            if (asyncio.isfuture(res) or inspect.isgenerator(res)):
                res = yield from res
            else:
                # If 'res' is an awaitable, run it.
                try:
                    await_meth = res.__await__
                except AttributeError:
                    pass
                else:
                    if isinstance(res, collections.abc.Awaitable):
                        res = yield from await_meth()
            return res

    coro = types.coroutine(coro)
    wrapper = coro
    wrapper._is_coroutine = asyncio.coroutines._is_coroutine  # For iscoroutinefunction().
    return wrapper




# ------------------------------------------------------------------------------
# This file is part of PyTango (http://pytango.rtfd.io)
#
# Copyright 2006-2012 CELLS / ALBA Synchrotron, Bellaterra, Spain
# Copyright 2013-2014 European Synchrotron Radiation Facility, Grenoble, France
#
# Distributed under the terms of the GNU Lesser General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
# ------------------------------------------------------------------------------

"""
This is the main PyTango package file.
Documentation for this package can be found online:

http://pytango.readthedocs.io
"""


__all__ = (
    'AccessControlType', 'ApiUtil', 'ArchiveEventInfo',
    'ArchiveEventProp', 'ArgType', 'AsynCall', 'AsynReplyNotArrived', 'AttReqType',
    'Attr', 'AttrConfEventData', 'AttrData', 'AttrDataFormat', 'AttrList',
    'AttrProperty', 'AttrQuality', 'AttrReadEvent', 'AttrSerialModel',
    'AttrWriteType', 'AttrWrittenEvent', 'Attribute', 'AttributeAlarm',
    'AttributeAlarmInfo', 'AttributeConfig', 'AttributeConfig_2',
    'AttributeConfig_3', 'AttributeDimension', 'AttributeEventInfo',
    'AttributeInfo', 'AttributeInfoEx', 'AttributeInfoList', 'AttributeInfoListEx',
    'AttributeList', 'AttributeProxy', 'ChangeEventInfo', 'ChangeEventProp',
    'Pipe', 'PipeConfig', 'PipeWriteType', 'PipeEventData', 'DevIntrChangeEventData',
    'CmdArgType', 'CmdDoneEvent', 'CommandInfo', 'CommandInfoList',
    'CommunicationFailed', 'Connection', 'ConnectionFailed',
    'ConstDevString', 'DServer', 'DataReadyEventData', 'Database', 'DbData',
    'DbDatum', 'DbDevExportInfo', 'DbDevExportInfos', 'DbDevImportInfo',
    'DbDevImportInfos', 'DbDevFullInfo', 'DbDevInfo', 'DbDevInfos', 'DbHistory',
    'DbHistoryList', 'DbServerInfo', 'DbServerData', 'DebugIt', 'DevBoolean', 'DevCommandInfo',
    'DevDouble', 'DevEncoded', 'DevError', 'DevFailed', 'DevFloat', 'DevInt',
    'DevLong', 'DevLong64', 'DevShort', 'DevSource', 'DevState', 'DevString',
    'DevUChar', 'DevULong', 'DevULong64', 'DevUShort', 'DevEnum', 'DevVarBooleanArray',
    'DevVarCharArray', 'DevVarDoubleArray', 'DevVarDoubleStringArray',
    'DevVarFloatArray', 'DevVarLong64Array', 'DevVarLongArray',
    'DevVarLongStringArray', 'DevVarShortArray', 'DevVarStringArray',
    'DevVarULong64Array', 'DevVarULongArray', 'DevVarUShortArray',
    'DevVoid', 'DeviceAttribute', 'DeviceAttributeConfig',
    'DeviceAttributeHistory', 'DeviceClass', 'DeviceData', 'DeviceDataList',
    'DeviceDataHistory', 'DeviceDataHistoryList',
    'DeviceImpl', 'DeviceInfo', 'DeviceProxy',
    'DeviceUnlocked', 'Device_2Impl', 'Device_3Impl', 'Device_4Impl', 'Device_5Impl',
    'DispLevel', 'EncodedAttribute', 'EnsureOmniThread', 'ErrSeverity', 'ErrorIt',
    'EventData', 'EventProperties', 'EventSystemFailed', 'EventType',
    'Except', 'ExtractAs', 'FMT_UNKNOWN', 'FatalIt', 'GreenMode', 'Group',
    'GroupAttrReply',
    'GroupAttrReplyList', 'GroupCmdReply', 'GroupCmdReplyList',
    'GroupReply', 'GroupReplyList', 'IMAGE', 'ImageAttr', 'InfoIt',
    'KeepAliveCmdCode', 'Level', 'LockCmdCode', 'LockerInfo', 'LockerLanguage',
    'LogIt', 'LogLevel', 'LogTarget', 'Logger', 'Logging', 'MessBoxType',
    'MultiAttribute', 'MultiAttrProp', 'MultiClassAttribute', 'NamedDevFailed',
    'NamedDevFailedList', 'NonDbDevice', 'NonSupportedFeature',
    'NotAllowed', 'PeriodicEventInfo', 'PeriodicEventProp',
    'PollCmdCode', 'PollDevice',
    'PollObjType', 'READ', 'READ_WITH_WRITE', 'READ_WRITE', 'Release', 'SCALAR',
    'SPECTRUM', 'SerialModel', 'SpectrumAttr', 'StdDoubleVector',
    'StdGroupAttrReplyVector', 'StdGroupCmdReplyVector', 'StdGroupReplyVector',
    'StdLongVector', 'StdNamedDevFailedVector', 'StdStringVector', 'SubDevDiag',
    'TangoStream', 'TimeVal', 'UserDefaultAttrProp', 'UserDefaultPipeProp', 'Util',
    'WAttribute', 'WRITE', 'WarnIt', 'WrongData', 'WrongNameSyntax', '__version__',
    '__version_description__', '__version_info__', '__version_long__',
    '__version_number__', 'alarm_flags', 'asyn_req_type', 'cb_sub_model',
    'class_factory', 'class_list', 'constants', 'constructed_class',
    'cpp_class_list', 'delete_class_list', 'get_class', 'get_classes',
    'get_constructed_class', 'get_constructed_classes', 'get_cpp_class',
    'get_cpp_classes', 'raise_asynch_exception', 'AutoTangoMonitor',
    'AutoTangoAllowThreads', 'LatestDeviceImpl', 'Interceptors',
    'get_attribute_proxy', 'requires_tango', 'requires_pytango',
    'set_green_mode', 'get_green_mode', 'get_device_proxy',
    'is_scalar_type', 'is_array_type', 'is_omni_thread', 'is_numerical_type',
    'is_int_type', 'is_float_type', 'is_bool_type', 'is_str_type',
    'obj_2_str', 'str_2_obj', 'seqStr_2_obj')

__docformat__ = "restructuredtext"

from ._tango import (
    AccessControlType, ApiUtil, ArchiveEventInfo,
    AsynCall, AsynReplyNotArrived, AttReqType, Attr, AttrConfEventData,
    AttrDataFormat, AttrList, AttrProperty, AttrQuality, AttrReadEvent,
    AttrSerialModel, AttrWriteType, AttrWrittenEvent, Attribute,
    AttributeAlarmInfo, AttributeDimension, AttributeEventInfo, AttributeInfo,
    AttributeInfoEx, AttributeInfoList, AttributeInfoListEx, AttributeList,
    ChangeEventInfo, CmdArgType, Pipe, PipeWriteType, DevIntrChangeEventData,
    CmdDoneEvent, CommandInfo, CommandInfoList, CommunicationFailed,
    Connection, ConnectionFailed, ConstDevString, DServer, DataReadyEventData,
    Database, DbData, DbDatum, DbDevExportInfo, DbDevExportInfos,
    DbDevImportInfo, DbDevImportInfos, DbDevFullInfo, DbDevInfo, DbDevInfos, DbHistory,
    DbHistoryList, DbServerInfo, DbServerData, DevBoolean, DevCommandInfo, DevDouble,
    DevEncoded, DevError, DevFailed, DevFloat, DevInt, DevLong, DevLong64,
    DevShort, DevSource, DevState, DevString, DevUChar, DevULong, DevULong64,
    DevUShort, DevEnum, DevVarBooleanArray, DevVarCharArray, DevVarDoubleArray,
    DevVarDoubleStringArray, DevVarFloatArray, DevVarLong64Array,
    DevVarLongArray, DevVarLongStringArray, DevVarShortArray, DevVarStringArray,
    DevVarULong64Array, DevVarULongArray, DevVarUShortArray, DevVoid,
    DeviceAttribute, DeviceAttributeConfig, DeviceAttributeHistory,
    DeviceData, DeviceDataList, DeviceDataHistory, DeviceDataHistoryList,
    DeviceImpl, DeviceInfo, DeviceProxy, DeviceUnlocked, Device_2Impl,
    Device_3Impl, Device_4Impl, Device_5Impl, DispLevel, EncodedAttribute,
    EnsureOmniThread, ErrSeverity, EventData, EventSystemFailed, EventType, PipeEventData,
    Except, ExtractAs, GreenMode, FMT_UNKNOWN, GroupAttrReply, GroupAttrReplyList,
    GroupCmdReply, GroupCmdReplyList, GroupReply, GroupReplyList,
    IMAGE, ImageAttr, KeepAliveCmdCode, Level, LockCmdCode, LockerInfo,
    LockerLanguage, LogLevel, LogTarget, Logger, Logging, MessBoxType,
    MultiAttribute, MultiClassAttribute, NamedDevFailed, NamedDevFailedList,
    NonDbDevice, NonSupportedFeature, NotAllowed, PeriodicEventInfo,
    PollCmdCode, PollDevice, PollObjType, READ,
    READ_WITH_WRITE, READ_WRITE, SCALAR, SPECTRUM, SerialModel,
    SpectrumAttr, StdDoubleVector, StdGroupAttrReplyVector,
    StdGroupCmdReplyVector, StdGroupReplyVector, StdLongVector,
    StdNamedDevFailedVector, StdStringVector, SubDevDiag, TimeVal,
    UserDefaultAttrProp, UserDefaultPipeProp, WAttribute, WRITE, WrongData,
    WrongNameSyntax, alarm_flags, asyn_req_type, cb_sub_model, constants,
    raise_asynch_exception, Interceptors,
    AutoTangoMonitor, AutoTangoAllowThreads, is_omni_thread)


# Aliases

ArgType = CmdArgType


# Release

from .release import Release

__author__ = Release.author_lines
__version_info__ = Release.version_info
__version_short__ = Release.version
__version__ = Release.version_long
__version_long__ = Release.version_long
__version_number__ = Release.version_number
__version_description__ = Release.version_description
__doc__ = Release.long_description

# Pytango imports

from .attr_data import AttrData

from .log4tango import (
    TangoStream, LogIt, DebugIt, InfoIt, WarnIt, ErrorIt, FatalIt)

from .device_server import (
    ChangeEventProp, PeriodicEventProp, ArchiveEventProp, AttributeAlarm,
    EventProperties, AttributeConfig, AttributeConfig_2, AttributeConfig_3,
    MultiAttrProp, LatestDeviceImpl)

from .pipe import PipeConfig

from .attribute_proxy import AttributeProxy, get_attribute_proxy

from .group import Group

from .pyutil import Util

from .device_class import DeviceClass

from .globals import (
    get_class, get_classes, get_cpp_class, get_cpp_classes,
    get_constructed_class, get_constructed_classes, class_factory,
    delete_class_list, class_list, cpp_class_list, constructed_class)

from .utils import (
    requires_pytango, requires_tango,
    is_scalar_type, is_array_type, is_numerical_type,
    is_int_type, is_float_type, is_bool_type, is_str_type,
    obj_2_str, str_2_obj, seqStr_2_obj)

from .green import set_green_mode, get_green_mode

from .device_proxy import get_device_proxy


# Pytango initialization

from .pytango_init import init as __init
__init()
requires_tango("9.4.1", software_name="PyTango")



from PyTango.server import Server

class MyClass:

    def __init__(self):
        self.bla = 55.6
        self.ble = 11
        self.bli = False

    def func1(self):
        return "executed func1"

    def func2(self, v):
        return 2*v

    def func3(self, a, b, c=1, d=3):
        """Just some documentation"""
        return "done func3"

    def func(self, nap_time):
        import time
        time.sleep(nap_time)
        return f"Finished sleep for {nap_time}s"

server = Server("many", server_type="Server")

N = 100

objs = []
for i in range(N):
    name = "obj%04d" % i
    obj = MyClass()
    objs.append(obj)
    server.register_object(obj, name)

server.run()



class ROI:
    def __init__(self, x, y, w, h):
        self.x = x
        self.y = y
        self.w = w
        self.h = h

    def __repr__(self):
        return str(self)

    def __str__(self):
        txt = f"ROI(x={self.x}, y={self.y}, w={self.w}, h={self.h})"
        return txt



"""Provide a context to run a device without a database."""


# Imports
import os
import sys
import time
import struct
import socket
import tempfile
import traceback
import collections
from functools import partial
import psutil

# Concurrency imports
import threading
import multiprocessing
import queue

# CLI imports
from ast import literal_eval
from importlib import import_module
from argparse import ArgumentParser, ArgumentTypeError

# Local imports
from .server import run
from .utils import is_non_str_seq
from . import DeviceProxy, Database, Util

__all__ = ("MultiDeviceTestContext", "DeviceTestContext", "run_device_test_context",
           "get_server_port_via_pid")

# Helpers

_WINDOWS = "nt" in os.name
if not _WINDOWS:
    _DEFAULT_THREAD_TIMEOUT = 3.
    _DEFAULT_PROCESS_TIMEOUT = 5.
else:
    _DEFAULT_THREAD_TIMEOUT = 5.
    _DEFAULT_PROCESS_TIMEOUT = 7.


IOR = collections.namedtuple(
    'IOR',
    'first dtype_length dtype nb_profile tag '
    'length major minor wtf host_length host port body')


def ascii_to_bytes(s):
    convert = lambda x: bytes((int(x, 16),))
    return b''.join(convert(s[i:i + 2]) for i in range(0, len(s), 2))


def parse_ior(encoded_ior):
    assert encoded_ior[:4] == 'IOR:'
    ior = ascii_to_bytes(encoded_ior[4:])
    dtype_length = struct.unpack_from('II', ior)[-1]
    form = f'II{dtype_length:d}sIIIBBHI'
    host_length = struct.unpack_from(form, ior)[-1]
    form = f'II{dtype_length:d}sIIIBBHI{host_length:d}sH0I'
    values = struct.unpack_from(form, ior)
    values += (ior[struct.calcsize(form):],)
    strip = lambda x: x[:-1] if isinstance(x, bytes) else x
    return IOR(*map(strip, values))


def get_server_host_port():
    util = Util.instance()
    ds = util.get_dserver_device()
    encoded_ior = util.get_dserver_ior(ds)
    ior = parse_ior(encoded_ior)
    return ior.host.decode(), ior.port


def get_server_port_via_pid(pid, host):
    """Return the TCP port that a device server process is listening on (GIOP).

    This checks TCP sockets open on the process with the given PID, and attempts
    to find the one that accepts GIOP traffic.  A connection will be made to each
    listening socket, and data may be sent to them.

    General Inter-ORB Protocol (GIOP) is the message protocol which object
    request brokers (ORBs) communicate in CORBA.  This port is the one that is
    used when connecting a DeviceProxy.  These are not the port(s) used for ZMQ
    event traffic.

    :param pid: operating system process identifier
    :type pid: int
    :param host: hostname/IP that device server is listening on.  E.g., 127.0.0.1,
    IP address of a non-loopback network interface, etc.  Note that starting a device
    server on "localhost" may fail if OmniORB creates an IPv6-only socket.
    :type pid: str

    :returns: TCP port number
    :rtype: int

    :raises RuntimeError: If the GIOP port couldn't be identified

    .. versionadded:: 9.4.0
    """
    ports = _get_listening_tcp_ports(pid)
    return _get_giop_port(host, ports)


def _get_listening_tcp_ports(pid):
    p = psutil.Process(pid)
    conns = p.connections(kind="tcp")
    return [c.laddr[1] for c in conns]


def _get_giop_port(host, ports):
    protocols = _try_get_protocols_on_ports(host, ports)
    for port, protocol in protocols.items():
        if protocol == "GIOP":
            return port
    raise RuntimeError(
        f"None of ports {ports} appear to have GIOP protocol. "
        f"Guessed protocols: {protocols}."
    )


def _try_get_protocols_on_ports(host, ports):
    """Return a dict with port to protocol mapping.

    This attempts to establish a TCP socket connection to the host
    for each port, and then determine the protocol it supports.

    ZMQ client sockets receive an unsolicited version check message on connection.
    CORBA GIOP client sockets don't receive an unsolicited message, so we send
    a requested to disconnect, and expect an empty message back.
    """
    zmq_response = b'\xff\x00\x00\x00\x00\x00\x00\x00\x01\x7f'  # signature + version check
    giop_send = b'GIOP\x01\x02\x01\x05\x00\x00\x00\x00'  # request disconnect
    giop_response = b''  # graceful disconnect
    max_bytes_expected = len(zmq_response)

    protocols = dict.fromkeys(ports, "Unknown")
    for port in ports:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.settimeout(0.001)
            server_address = (host, port)
            sock.connect(server_address)

            try:
                data = sock.recv(max_bytes_expected)
                if data == zmq_response:
                    protocols[port] = "ZMQ"
                    continue
            except OSError:
                pass

            try:
                sock.sendall(giop_send)
                data = sock.recv(max_bytes_expected)
                if data == giop_response:
                    protocols[port] = "GIOP"
                    continue
            except OSError:
                pass
        except OSError:
            pass
        finally:
            sock.close()
    return protocols


def literal_dict(arg):
    return dict(literal_eval(arg))


def device(path):
    """Get the device class from a given module."""
    module_name, device_name = path.rsplit(".", 1)
    try:
        module = import_module(module_name)
    except Exception:
        raise ArgumentTypeError(f"Error importing {module_name}.{device_name}:\n"
                                f"{traceback.format_exc()}")
    return getattr(module, device_name)


def get_host_ip():
    """Get the primary external host IP.

    This is useful because an explicit IP is required to get
    tango events to work properly. Note that localhost does not work
    either.
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    # Connecting to a UDP address doesn't send packets
    # Note:  For some reason Python3 on macOS does not accept 0 as a port but
    # returns with an errno 49 instead. Therefore just use port 80 which just
    # works as well.
    s.connect(('8.8.8.8', 80))
    # Get ip address
    ip = s.getsockname()[0]
    return ip


def _device_class_from_field(field):
    """
    Helper function that extracts and return a device class from a
    'class' field of a devices_info dictionary

    :param field: the field from which to extract the device class

    :return: the device class extracted from the field
    """
    device_cls_class = None
    if is_non_str_seq(field):
        (device_cls_class, device_class) = (field[0], field[1])
        if isinstance(device_cls_class, str):
            device_cls_class = device(device_cls_class)
    else:
        device_class = field
    if isinstance(device_class, str):
        device_class = device(device_class)
    return (device_cls_class, device_class)


class MultiDeviceTestContext:
    """Context to run device(s) without a database.

    The difference with respect to
    :class:`~tango.test_context.DeviceTestContext` is that it allows
    to export multiple devices (even of different Tango classes).

    Example usage::

        from tango.server import Device, attribute
        from tango.test_context import MultiDeviceTestContext


        class Device1(Device):

            @attribute
            def attr1(self):
                return 1.0


        class Device2(Device):

            @attribute
            def read_attr2(self):
                return 2.0


        devices_info = (
            {
                "class": Device1,
                "devices": [
                    {
                        "name": "test/device1/1"
                    },
                ]
            },
            {
                "class": Device2,
                "devices": [
                    {
                        "name": "test/device2/1",
                    },
                ]
            }
        )

        def test_devices():
            with MultiDeviceTestContext(devices_info, process=True) as context:
                proxy1 = context.get_device("test/device1/1")
                proxy2 = context.get_device("test/device2/1")
                assert proxy1.attr1 == 1.0
                assert proxy2.attr2 == 2.0

    :param devices_info:
      a sequence of dicts with information about
      devices to be exported. Each dict consists of the following keys:

        * "class" which value is either of:

          * :class:`~tango.server.Device` or the name of some such class
          * a sequence of two elements, the first element being a
            :class:`~tango.DeviceClass` or the name of some such class,
            the second element being a :class:`~tango.DeviceImpl` or the
            name of such such class

        * "devices" which value is a sequence of dicts with the following keys:

          * "name" (str)
          * "properties" (dict)
          * "memorized" (dict)

    :type devices_info:
      sequence<dict>
    :param server_name:
      Name to use for the device server.
      Optional.  Default is the first device's class name.
    :type server_name:
      :py:obj:`str`
    :param instance_name:
      Name to use for the device server instance.
      Optional.  Default is lower-case version of the server name.
    :type instance_name:
      :py:obj:`str`
    :param db:
      Path to a pre-populated text file to use for the
      database.
      Optional.  Default is to create a new temporary file and populate it
      based on the devices and properties supplied in `devices_info`.
    :type db:
      :py:obj:`str`
    :param host:
      Hostname to use for device server's ORB endpoint.
      Optional.  Default is a local IP address.
    :type host:
      :py:obj:`str`
    :param port:
      Port number to use for the device server's ORB endpoint.
      Optional.  Default is chosen by omniORB.
    :type port:
      :py:obj:`int`
    :param debug:
      Debug level for the device server logging.
      0=OFF, 1=FATAL, 2=ERROR, 3=WARN, 4=INFO, 5=DEBUG.
      Optional. Default is warn.
    :type debug:
      :py:obj:`int`
    :param process:
      True if the device server should be launched in a new process, otherwise
      use a new thread.  Note:  if the context will be used mutiple times, it
      may seg fault if the thread mode is chosen.
      Optional.  Default is thread.
    :type process:
      :py:obj:`bool`
    :param daemon:
      True if the new thread/process must be created in daemon mode.
      Optional.  Default is not daemon.
    :type daemon:
      :py:obj:`bool`
    :param timeout:
      How long to wait (seconds) for the device server to start up, and also
      how long to wait on joining the thread/process when stopping.
      Optional.  Default differs for thread and process modes.
    :type timeout:
      :py:obj:`float`
    :param green_mode:
      Green mode to use for the device server.
      Optional.  Default uses the Device specification (via green_mode class attribute),
      or if that isn't specified the global green mode.
    :type green_mode:
      :obj:`~tango.GreenMode`
    """
    nodb = "dbase=no"
    command = "{0} {1} -ORBendPoint giop:tcp:{2}:{3} -file={4}"

    thread_timeout = _DEFAULT_THREAD_TIMEOUT
    process_timeout = _DEFAULT_PROCESS_TIMEOUT

    def __init__(self, devices_info, server_name=None, instance_name=None,
                 db=None, host=None, port=0, debug=3,
                 process=False, daemon=False, timeout=None, green_mode=None):
        if not server_name:
            _, first_device = _device_class_from_field(devices_info[0]["class"])
            server_name = first_device.__name__
        if not instance_name:
            instance_name = server_name.lower()
        if db is None:
            handle, db = tempfile.mkstemp()
            self.handle = handle
        else:
            self.handle = None
        if host is None:
            # IP address is used instead of the hostname on purpose (see #246)
            host = get_host_ip()
        if timeout is None:
            timeout = self.process_timeout if process else self.thread_timeout
        # Patch bug #819
        if process:
            os.environ['ORBscanGranularity'] = '0'
        # Attributes
        self.db = db
        self.host = host
        self.port = port
        self.timeout = timeout
        self.server_name = "/".join(("dserver", server_name, instance_name))
        self.queue = multiprocessing.Queue() if process else queue.Queue()
        self._devices = {}

        # Command args
        string = self.command.format(
            server_name, instance_name, host, port, db)
        string += f" -v{debug}" if debug else ""
        cmd_args = string.split()

        class_list = []
        device_list = []
        tangoclass_list = []
        for device_info in devices_info:
            device_cls, device = _device_class_from_field(device_info["class"])
            tangoclass = device.__name__
            if tangoclass in tangoclass_list:
                self.delete_db()
                raise ValueError("multiple entries in devices_info pointing "
                                 "to the same Tango class")
            tangoclass_list.append(tangoclass)
            # File
            self.append_db_file(server_name, instance_name, tangoclass,
                                device_info["devices"])
            if device_cls:
                class_list.append((device_cls, device, tangoclass))
            else:
                device_list.append(device)

        # Target and arguments
        if class_list and device_list:
            self.delete_db()
            raise ValueError("mixing HLAPI and classical API in devices_info "
                             "is not supported")
        if class_list:
            runserver = partial(run, class_list, cmd_args, green_mode=green_mode)
        elif len(device_list) == 1 and hasattr(device_list[0], "run_server"):
            runserver = partial(device.run_server, cmd_args, green_mode=green_mode)
        elif device_list:
            runserver = partial(run, device_list, cmd_args, green_mode=green_mode)
        else:
            raise ValueError("Wrong format of devices_info")

        cls = multiprocessing.Process if process else threading.Thread
        self.thread = cls(target=self.target, args=(runserver, process))
        self.thread.daemon = daemon

    def target(self, runserver, process=False):
        try:
            runserver(post_init_callback=self.post_init, raises=True)
        except Exception:
            # Put exception in the queue
            etype, value, tb = sys.exc_info()
            if process:
                tb = None  # Traceback objects can't be pickled
            self.queue.put((etype, value, tb))
        finally:
            # Put something in the queue just in case
            exc = RuntimeError("The server failed to report anything")
            self.queue.put((None, exc, None))
            # Make sure the process has enough time to send the items
            # because the it might segfault while cleaning up the
            # the tango resources
            if process:
                time.sleep(0.1)

    def post_init(self):
        try:
            host, port = get_server_host_port()
            self.queue.put((host, port))
        except Exception as exc:
            self.queue.put((None, exc, None))
        finally:
            # Put something in the queue just in case
            exc = RuntimeError(
                "The post_init routine failed to report anything")
            self.queue.put((None, exc, None))

    def append_db_file(self, server, instance, tangoclass, device_prop_info):
        """Generate a database file corresponding to the given arguments.
        """
        device_names = [info["name"] for info in device_prop_info]
        # Open the file
        with open(self.db, "a") as f:
            f.write("/".join((server, instance, "DEVICE", tangoclass)))
            f.write(": ")
            f.write(", ".join(device_names))
            f.write("\n")
        # Create database
        db = Database(self.db)
        # Write properties
        for info in device_prop_info:
            device_name = info["name"]
            properties = info.get("properties", {})
            # Patch the property dict to avoid a PyTango bug
            patched = {key: value if value != '' else ' '
                           for key, value in properties.items()}
            db.put_device_property(device_name, patched)

            memorized = info.get("memorized", {})
            munged = {
                attribute_name: {
                    "__value": memorized_value
                } for (attribute_name, memorized_value) in memorized.items()
            }
            db.put_device_attribute_property(device_name, munged)
        return db

    def delete_db(self):
        """ delete temporary database file only if it was created by this class """
        if self.handle is not None:
            os.close(self.handle)
            os.unlink(self.db)

    def get_server_access(self):
        """Return the full server name."""
        return f'tango://{self.host}:{self.port}/{self.server_name}#{self.nodb}'

    def get_device_access(self, device_name):
        """Return the full device name."""
        return f'tango://{self.host}:{self.port}/{device_name}#{self.nodb}'

    def get_device(self, device_name):
        """Return the device proxy corresponding to the given device name.

        Maintains previously accessed device proxies in a cache to not recreate
        then on every access.
        """
        if device_name not in self._devices:
            device = DeviceProxy(self.get_device_access(device_name))
            self._devices[device_name] = device
        return self._devices[device_name]

    def start(self):
        """Run the server."""
        self.thread.start()
        self.connect()
        return self

    def connect(self):
        try:
            args = self.queue.get(timeout=self.timeout)
        except queue.Empty:
            if self.thread.is_alive():
                raise RuntimeError(
                    'The server appears to be stuck at initialization. '
                    'Check stdout/stderr for more information.')
            elif hasattr(self.thread, 'exitcode'):
                raise RuntimeError(
                    f'The server process stopped with exitcode {self.thread.exitcode}. '
                    f'Check stdout/stderr for more information.')
            else:
                raise RuntimeError(
                    'The server stopped without reporting. '
                    'Check stdout/stderr for more information.')
        try:
            self.host, self.port = args
        except ValueError as e:
            # Warning: an assumption made here - if the args don't contain the
            #          expected (host, port) items then we assume the args contain
            #          exception info.
            if len(args) == 3:
                exc_type, exc_value, exc_traceback = args
                raise exc_value
            # In case the above assumption is wrong: re-raise the original exception
            raise ValueError(*args) from e
        # Get server proxy
        self.server = DeviceProxy(self.get_server_access())
        self.server.ping()

    def stop(self):
        """Kill the server."""
        try:
            if self.server:
                self.server.command_inout('Kill')
            self.join(self.timeout)
        finally:
            self.delete_db()

    def join(self, timeout=None):
        self.thread.join(timeout)

    def __enter__(self):
        """Enter method for context support.

        :return:
          Instance of this test context.  Use `get_device` to get proxy
          access to any of the devices started by this context.
        :rtype:
          :class:`~tango.test_context.MultiDeviceTestContext`
        """
        if not self.thread.is_alive():
            self.start()
        return self

    def __exit__(self, exc_type, exception, trace):
        """Exit method for context support."""
        self.stop()
        return False


# Single device test context
class DeviceTestContext(MultiDeviceTestContext):
    """Context to run a single device without a database.

    The difference with respect to
    :class:`~tango.test_context.MultiDeviceTestContext` is that it only
    allows to export a single device.

    Example usage::

        from time import sleep

        from tango.server import Device, attribute, command
        from tango.test_context import DeviceTestContext

        class PowerSupply(Device):

            @attribute(dtype=float)
            def voltage(self):
                return 1.23

            @command
            def calibrate(self):
                sleep(0.1)

        def test_calibrate():
            '''Test device calibration and voltage reading.'''
            with DeviceTestContext(PowerSupply, process=True) as proxy:
                proxy.calibrate()
                assert proxy.voltage == 1.23

    :param device:
      Device class to be run.
    :type device:
      :class:`~tango.server.Device` or :class:`~tango.DeviceImpl`
    :param device_cls:
      The device class can be provided if using the low-level API.
      Optional.  Not required for high-level API devices, of type
      :class:`~tango.server.Device`.
    :type device_cls:
      :class:`~tango.DeviceClass`

     The rest of the parameters are described in
     :class:`~tango.test_context.MultiDeviceTestContext`.
     """

    def __init__(self, device, device_cls=None, server_name=None,
                 instance_name=None, device_name=None, properties=None,
                 db=None, host=None, port=0, debug=3, process=False,
                 daemon=False, timeout=None, memorized=None, green_mode=None):
        # Argument
        if not server_name:
            server_name = device.__name__
        if not instance_name:
            instance_name = server_name.lower()
        if not device_name:
            device_name = 'test/nodb/' + server_name.lower()
        if properties is None:
            properties = {}
        if memorized is None:
            memorized = {}
        if device_cls:
            cls = (device_cls, device)
        else:
            cls = device
        devices_info = (
            {
                "class": cls,
                "devices": (
                    {
                        "name": device_name,
                        "properties": properties,
                        "memorized": memorized},
                )
            },
        )
        super().__init__(devices_info,
                                                server_name=server_name,
                                                instance_name=instance_name,
                                                db=db, host=host,
                                                port=port, debug=debug,
                                                process=process,
                                                daemon=daemon,
                                                timeout=timeout,
                                                green_mode=green_mode)

        self.device_name = device_name
        self.device = self.server = None

    def get_device_access(self, device_name=None):
        """Return the full device name."""
        if device_name is None:
            device_name = self.device_name
        return super().get_device_access(
            device_name)

    def connect(self):
        super().connect()
        # Get device proxy
        self.device = self.get_device(self.device_name)
        self.device.ping()

    def __enter__(self):
        """Enter method for context support.

        :return:
          A device proxy to the device started by this context.
        :rtype:
          :class:`~tango.DeviceProxy`
        """
        if not self.thread.is_alive():
            self.start()
        return self.device


# Command line interface

def parse_command_line_args(args=None):
    """Parse arguments given in command line."""
    desc = "Run a given device on a given port."
    parser = ArgumentParser(description=desc)
    # Add arguments
    msg = 'The device to run as a python path.'
    parser.add_argument('device', metavar='DEVICE',
                        type=device, help=msg)
    msg = "The hostname to use."
    parser.add_argument('--host', metavar='HOST',
                        type=str, help=msg, default=None)
    msg = "The port to use."
    parser.add_argument('--port', metavar='PORT',
                        type=int, help=msg, default=8888)
    msg = "The debug level."
    parser.add_argument('--debug', metavar='DEBUG',
                        type=int, help=msg, default=3)
    msg = "The properties to set as python dict."
    parser.add_argument('--prop', metavar='PROP',
                        type=literal_dict, help=msg, default='{}')
    # Parse arguments
    namespace = parser.parse_args(args)
    return (namespace.device, namespace.host, namespace.port,
            namespace.prop, namespace.debug)


def run_device_test_context(args=None):
    device, host, port, properties, debug = parse_command_line_args(args)
    context = DeviceTestContext(
        device, properties=properties, host=host, port=port, debug=debug)
    context.start()
    msg = f'{device.__name__} started on port {context.port} with properties {properties}'
    print(msg)
    print(f'Device access: {context.get_device_access()}')
    print(f'Server access: {context.get_server_access()}')
    context.join()
    print("Done")


# Main execution

if __name__ == "__main__":
    run_device_test_context()



"""Client tests that run against the standard TangoTest device.

Due to a TANGO 9 issue (#821), the device is run without any database.
Note that this means that various features won't work:

 * No device configuration via properties.
 * No event generated by the server.
 * No memorized attributes.
 * No device attribute configuration via the database.

So don't even try to test anything of the above as it will not work
and is even likely to crash the device (!)

"""

import gc
import weakref

from distutils.spawn import find_executable
from subprocess import Popen
from time import sleep

import pytest
from functools import partial
from tango import DeviceProxy, DevFailed, GreenMode
from tango import DeviceInfo, AttributeInfo, AttributeInfoEx, ExtractAs
from tango.server import Device
from tango.utils import is_str_type, is_int_type, is_float_type, is_bool_type
from tango.test_utils import (
    DeviceTestContext,
    assert_close,
    bytes_devstring,
    convert_to_type,
    get_server_port_via_pid,
    str_devstring,
)
from tango.gevent import DeviceProxy as gevent_DeviceProxy
from tango.futures import DeviceProxy as futures_DeviceProxy
from tango.asyncio import DeviceProxy as asyncio_DeviceProxy

# Asyncio imports
try:
    import asyncio
except ImportError:
    import trollius as asyncio  # noqa: F401


ATTRIBUTES = [
              'State',
              'Status',
              'ampli',
              'boolean_image',
              'boolean_image_ro',
              'boolean_scalar',
              'boolean_spectrum',
              'boolean_spectrum_ro',
              'double_image',
              'double_image_ro',
              'double_scalar',
              'double_scalar_rww',
              'double_scalar_w',
              'double_spectrum',
              'double_spectrum_ro',
              'enum_image',
              'enum_image_ro',
              'enum_scalar',
              'enum_scalar_ro',
              'enum_spectrum',
              'enum_spectrum_ro',
              'float_image',
              'float_image_ro',
              'float_scalar',
              'float_spectrum',
              'float_spectrum_ro',
              'freq',
              'long64_image_ro',
              'long64_scalar',
              'long64_spectrum_ro',
              'long_image',
              'long_image_ro',
              'long_scalar',
              'long_scalar_rww',
              'long_scalar_w',
              'long_spectrum',
              'long_spectrum_ro',
              'no_value',
              'short_image',
              'short_image_ro',
              'short_scalar',
              'short_scalar_ro',
              'short_scalar_rww',
              'short_scalar_w',
              'short_spectrum',
              'short_spectrum_ro',
              'string_image',
              'string_image_ro',
              'string_scalar',
              'string_spectrum',
              'string_spectrum_ro',
              'throw_exception',
              'uchar_image',
              'uchar_image_ro',
              'uchar_scalar',
              'uchar_spectrum',
              'uchar_spectrum_ro',
              'ulong64_image_ro',
              'ulong64_scalar',
              'ulong64_spectrum_ro',
              'ulong_image_ro',
              'ulong_scalar',
              'ulong_spectrum_ro',
              'ushort_image',
              'ushort_image_ro',
              'ushort_scalar',
              'ushort_spectrum',
              'ushort_spectrum_ro',
              'wave'
             ]


device_proxy_map = {
    GreenMode.Synchronous: DeviceProxy,
    GreenMode.Futures: futures_DeviceProxy,
    GreenMode.Asyncio: partial(asyncio_DeviceProxy, wait=True),
    GreenMode.Gevent: gevent_DeviceProxy}


# Helpers

def start_server(host, server, inst, device):
    exe = find_executable(server)
    cmd = f"{exe} {inst} -ORBendPoint giop:tcp:{host}:0 -nodb -dlist {device}"
    proc = Popen(cmd.split(), close_fds=True)
    proc.poll()
    return proc


def get_proxy(host, port, device, green_mode):
    access = f"tango://{host}:{port}/{device}#dbase=no"
    return device_proxy_map[green_mode](access)


def wait_for_proxy(host, proc, device, green_mode, retries=400, delay=0.01):
    last_error = None
    count = 0
    port = None
    while port is None and count < retries:
        try:
            port = get_server_port_via_pid(proc.pid, host)
        except RuntimeError as exc:
            last_error = str(exc)
            sleep(delay)
        count += 1

    if port is not None:
        count = 0
        while count < retries:
            try:
                proxy = get_proxy(host, port, device, green_mode)
                proxy.ping()
                proxy.state()
                return proxy
            except DevFailed as exc:
                last_error = str(exc)
                sleep(delay)
            count += 1
    raise RuntimeError(
        f"TangoTest device did not start up within {retries * delay:.1f} sec!\n"
        f"Last error: {last_error}."
    )


def ping_device(proxy):
    if proxy.get_green_mode() == GreenMode.Asyncio:
        asyncio.get_event_loop().run_until_complete(proxy.ping())
    else:
        proxy.ping()


# Fixtures

@pytest.fixture(params=[GreenMode.Synchronous,
                        GreenMode.Asyncio,
                        GreenMode.Gevent,
                        GreenMode.Futures],
                scope="module")
def tango_test_with_green_modes(request):
    green_mode = request.param
    server = "TangoTest"
    inst = "test"
    device = "sys/tg_test/17"
    host = "127.0.0.1"
    proc = start_server(host, server, inst, device)
    proxy = wait_for_proxy(host, proc, device, green_mode)

    yield proxy

    proc.terminate()
    # let's not wait for it to exit, that takes too long :)


@pytest.fixture(scope="module")
def tango_test():
    green_mode = GreenMode.Synchronous
    server = "TangoTest"
    inst = "test"
    device = "sys/tg_test/17"
    host = "127.0.0.1"
    proc = start_server(host, server, inst, device)
    proxy = wait_for_proxy(host, proc, device, green_mode)

    yield proxy

    proc.terminate()


@pytest.fixture(params=ATTRIBUTES)
def attribute(request):
    return request.param


@pytest.fixture(params=[a for a in ATTRIBUTES
                        if a not in ("no_value", "throw_exception")])
def readable_attribute(request):
    return request.param


@pytest.fixture(params=[a for a in ATTRIBUTES
                        if "scalar" in a and
                        a.split("_")[-1] not in ("ro", "rww")])
def writable_scalar_attribute(request):
    return request.param


@pytest.fixture(params=[a for a in ATTRIBUTES
                        if "spectrum" in a and
                        a.split("_")[-1] not in ("ro", "rww")])
def writable_spectrum_attribute(request):
    return request.param


@pytest.fixture(params=['double_scalar_w', 'double_spectrum', 'double_image'])
def all_double_writable_attributes(request):
    return request.param


@pytest.fixture(params=[GreenMode.Synchronous,
                        GreenMode.Asyncio,
                        GreenMode.Gevent,
                        GreenMode.Futures])
def green_mode_device_proxy(request):
    green_mode = request.param
    return device_proxy_map[green_mode]


@pytest.fixture
def simple_device_fqdn():
    class TestDevice(Device):
        pass

    context = DeviceTestContext(TestDevice, host="127.0.0.1")
    context.start()
    yield context.get_device_access()
    context.stop()


# Tests

def test_ping(tango_test):
    duration = tango_test.ping(wait=True)
    assert isinstance(duration, int)


def test_info(tango_test):
    info = tango_test.info()
    assert isinstance(info, DeviceInfo)
    assert info.dev_class == "TangoTest"
    # ...


def test_read_attribute(tango_test, readable_attribute):
    "Check that readable attributes can be read"
    # For read-only string spectrum and read-only string image types,
    # the following error is very likely to be raised:
    # -> MARSHAL CORBA system exception: MARSHAL_PassEndOfMessage
    # An explicit sleep fixes the problem, but it's annoying to maintain

    if readable_attribute in ["string_image_ro", "string_spectrum_ro"]:
        pytest.xfail()
    tango_test.read_attribute(readable_attribute, wait=True)


def test_read_write_attribute_with_green_modes(tango_test_with_green_modes, all_double_writable_attributes):
    """
    Check that attributes can be read/write with all green modes
    """
    attr_name = all_double_writable_attributes
    if attr_name == 'double_scalar_w':
        write_values = -28.2
    elif attr_name == 'double_spectrum':
        write_values = [-28.2, 23.4]
    else:
        write_values = [[-28.2, 23.4], [-4.9, 6.5]]

    tango_test_with_green_modes.write_attribute(attr_name, write_values, wait=True)
    read_attr = tango_test_with_green_modes.read_attribute(attr_name, wait=True)
    assert_close(read_attr.value, write_values)
    assert_close(read_attr.w_value, write_values)


def test_write_scalar_attribute(tango_test, writable_scalar_attribute):
    "Check that writable scalar attributes can be written"
    attr_name = writable_scalar_attribute
    config = tango_test.get_attribute_config(attr_name, wait=True)
    if is_bool_type(config.data_type):
        tango_test.write_attribute(attr_name, True, wait=True)
    elif is_int_type(config.data_type):
        tango_test.write_attribute(attr_name, 76, wait=True)
    elif is_float_type(config.data_type):
        tango_test.write_attribute(attr_name, -28.2, wait=True)
    elif is_str_type(config.data_type):
        tango_test.write_attribute(attr_name, "hello", wait=True)
    else:
        pytest.xfail("Not currently testing this type")


def test_write_read_spectrum_attribute(tango_test, writable_spectrum_attribute, extract_as):
    "Check that writable spectrum attributes can be written and read"
    requested_type, expected_type = extract_as
    attr_name = writable_spectrum_attribute
    config = tango_test.get_attribute_config(attr_name, wait=True)
    if is_bool_type(config.data_type):
        write_values = [True, False]
    elif is_int_type(config.data_type):
        write_values = [76, 77]
    elif is_float_type(config.data_type):
        if requested_type == ExtractAs.String:
            # it is hard to find a proper float values, which can be converted to str,
            # most of them case UnicodeDecodeError, so we use the most simple one
            write_values = [0, 0]
        else:
            write_values = [-28.2, 44.3]
    elif is_str_type(config.data_type):
        if requested_type == ExtractAs.Numpy:
            expected_type = tuple
        if requested_type in [ExtractAs.ByteArray, ExtractAs.Bytes, ExtractAs.String]:
            pytest.xfail('Conversion from (str,) to ByteArray, Bytes and String not supported. May be fixed in future')
        write_values = ["hello", "hola"]
    else:
        pytest.xfail("Not currently testing this type")

    tango_test.write_attribute(attr_name, write_values, wait=True)
    read_attr = tango_test.read_attribute(attr_name, extract_as=requested_type)

    assert isinstance(read_attr.value, expected_type)
    assert_close(read_attr.value, convert_to_type(write_values, config.data_type, expected_type))

    assert isinstance(read_attr.w_value, expected_type)
    assert_close(read_attr.w_value, convert_to_type(write_values, config.data_type, expected_type))


def test_write_read_empty_spectrum_attribute(tango_test, writable_spectrum_attribute):
    "Check that writing empty list to spectrum attribute reads back as None."
    attr_name = writable_spectrum_attribute
    config = tango_test.get_attribute_config(attr_name, wait=True)
    if is_str_type(config.data_type):
        pytest.xfail('Conversion from (str,) to numpy not supported. Probably, may be fixed in future')

    tango_test.write_attribute(attr_name, [], wait=True)
    read_attr = tango_test.read_attribute(attr_name, wait=True)
    assert not len(read_attr.value)


def test_write_read_string_attribute(tango_test):
    attr_name = 'string_scalar'
    bytes_big = 100000 * b'big data '
    str_big = bytes_big.decode('latin-1')

    values = [b'', '', 'Hello, World!', b'Hello, World!',
              bytes_devstring, str_devstring, bytes_big, str_big]

    expected_values = ['', '', 'Hello, World!', 'Hello, World!',
                       str_devstring, str_devstring,
                       str_big, str_big]

    for value, expected_value in zip(values, expected_values):
        tango_test.write_attribute(attr_name, value, wait=True)
        result = tango_test.read_attribute(attr_name, wait=True)
        assert result.value == expected_value

    attr_name = 'string_spectrum'
    for value, expected_value in zip(values, expected_values):
        tango_test.write_attribute(attr_name, ['', value, ''], wait=True)
        result = tango_test.read_attribute(attr_name, wait=True)
        assert result.value[1] == expected_value

    attr_name = 'string_image'
    for value, expected_value in zip(values, expected_values):
        tango_test.write_attribute(attr_name, [[value], [value]], wait=True)
        result = tango_test.read_attribute(attr_name, wait=True)
        assert result.value == ((expected_value,), (expected_value,))


def test_set_non_existent_attribute_raises_by_default(tango_test):
    with pytest.raises(AttributeError, match="some_invalid_name"):
        tango_test.some_invalid_name = "123"


def test_set_non_existent_attribute_allowed_if_dynamic_interface_unfrozen(tango_test):
    tango_test.unfreeze_dynamic_interface()
    tango_test.some_invalid_name = "123"
    assert tango_test.some_invalid_name == "123"


def test_dynamic_interface_can_be_toggled(tango_test):
    tango_test.unfreeze_dynamic_interface()
    tango_test.some_invalid_name = "456"
    assert tango_test.some_invalid_name == "456"
    tango_test.freeze_dynamic_interface()
    with pytest.raises(AttributeError, match="another_invalid_name"):
        tango_test.another_invalid_name = "123"


def test_dynamic_interface_flag_can_be_read(tango_test):
    tango_test.unfreeze_dynamic_interface()
    assert not tango_test.is_dynamic_interface_frozen()
    tango_test.freeze_dynamic_interface()
    assert tango_test.is_dynamic_interface_frozen()


def test_dynamic_interface_only_applies_to_device_proxy_instance(tango_test):
    other_proxy = DeviceProxy(tango_test.adm_name())
    tango_test.unfreeze_dynamic_interface()
    other_proxy.freeze_dynamic_interface()
    assert not tango_test.is_dynamic_interface_frozen()
    assert other_proxy.is_dynamic_interface_frozen()


def test_dynamic_interface_unfreeze_generates_a_user_warning(tango_test):
    with pytest.warns(UserWarning):
        tango_test.unfreeze_dynamic_interface()


def test_read_attribute_config(tango_test, attribute):
    tango_test.get_attribute_config(attribute)


def test_read_attribute_config_with_green_modes(tango_test_with_green_modes, all_double_writable_attributes):
    tango_test_with_green_modes.get_attribute_config(all_double_writable_attributes)


def test_attribute_list_query(tango_test_with_green_modes):
    attrs = tango_test_with_green_modes.attribute_list_query()
    assert all(isinstance(a, AttributeInfo) for a in attrs)
    assert {a.name for a in attrs} == set(ATTRIBUTES)


def test_attribute_list_query_ex(tango_test_with_green_modes):
    attrs = tango_test_with_green_modes.attribute_list_query_ex()
    assert all(isinstance(a, AttributeInfoEx) for a in attrs)
    assert {a.name for a in attrs} == set(ATTRIBUTES)


def test_device_proxy_dir_method(tango_test):
    lst = dir(tango_test)
    attrs = tango_test.get_attribute_list()
    cmds = tango_test.get_command_list()
    pipes = tango_test.get_pipe_list()
    methods = dir(type(tango_test))
    internals = tango_test.__dict__.keys()
    # Check attributes
    assert set(attrs) < set(lst)
    assert set(map(str.lower, attrs)) < set(lst)
    # Check commands
    assert set(cmds) < set(lst)
    assert set(map(str.lower, cmds)) < set(lst)
    # Check pipes
    assert set(pipes) < set(lst)
    assert set(map(str.lower, pipes)) < set(lst)
    # Check internals
    assert set(methods) <= set(lst)
    # Check internals
    assert set(internals) <= set(lst)


def test_device_polling_command(tango_test):

    dct = {"SwitchStates": 1000, "DevVoid": 10000, "DumpExecutionState": 5000}

    for command, period in dct.items():
        tango_test.poll_command(command, period)

    ans = tango_test.polling_status()
    for info in ans:
        lines = info.split('\n')
        command = lines[0].split('= ')[1]
        period = int(lines[1].split('= ')[1])
        assert dct[command] == period


def test_device_polling_attribute(tango_test):

    dct = {"boolean_scalar": 1000, "double_scalar": 10000, "long_scalar": 5000}

    for attr, poll_period in dct.items():
        tango_test.poll_attribute(attr, poll_period)

    ans = tango_test.polling_status()
    for x in ans:
        lines = x.split('\n')
        attr = lines[0].split('= ')[1]
        poll_period = int(lines[1].split('= ')[1])
        assert dct[attr] == poll_period


def test_command_string(tango_test):
    cmd_name = 'DevString'
    bytes_big = 100000 * b'big data '
    str_big = bytes_big.decode('latin-1')

    values = [b'', '', 'Hello, World!', b'Hello, World!',
              bytes_devstring, str_devstring, bytes_big, str_big]

    expected_values = ['', '', 'Hello, World!', 'Hello, World!',
                       str_devstring, str_devstring,
                       str_big, str_big]

    for value, expected_value in zip(values, expected_values):
        result = tango_test.command_inout(cmd_name, value, wait=True)
        assert result == expected_value

    cmd_name = 'DevVarStringArray'
    for value, expected_value in zip(values, expected_values):
        result = tango_test.command_inout(cmd_name, [value, value], wait=True)
        assert result == [expected_value, expected_value]

    cmd_name = 'DevVarLongStringArray'
    for value, expected_value in zip(values, expected_values):
        result = tango_test.command_inout(cmd_name,
                                          [[-10, 200], [value, value]],
                                          wait=True)
        assert len(result) == 2
        assert_close(result[0], [-10, 200])
        assert_close(result[1], [expected_value, expected_value])


def test_repr_uses_info(green_mode_device_proxy, simple_device_fqdn):
    proxy = green_mode_device_proxy(simple_device_fqdn)
    assert repr(proxy) == 'TestDevice(test/nodb/testdevice)'


def test_repr_default_if_info_unavailable(green_mode_device_proxy):
    proxy = green_mode_device_proxy(
        "tango://localhost:0/invalid/test/dev#dbase=no"
    )

    def bad_info(self):
        raise RuntimeError("Break info for test")

    proxy.__class__.info = bad_info

    assert repr(proxy) == 'Device(invalid/test/dev)'


def test_multiple_repr_calls_only_call_info_once(
        green_mode_device_proxy, simple_device_fqdn):
    proxy = green_mode_device_proxy(simple_device_fqdn)

    def mock_info(self):
        self.info_call_count += 1
        return self.info_orig()

    proxy.__class__.info_orig = proxy.info
    proxy.__class__.info = mock_info
    proxy.__dict__["info_call_count"] = 0

    repr(proxy)
    assert proxy.info_call_count == 1
    repr(proxy)
    assert proxy.info_call_count == 1


def test_no_memory_leak_for_repr(green_mode_device_proxy, simple_device_fqdn):
    proxy = green_mode_device_proxy(simple_device_fqdn)
    ping_device(proxy)
    weak_ref = weakref.ref(proxy)

    repr(proxy)

    # clear strong reference and check if object can be garbage collected
    del proxy
    assert_object_released_after_gc(weak_ref)


def test_no_memory_leak_for_str(green_mode_device_proxy, simple_device_fqdn):
    proxy = green_mode_device_proxy(simple_device_fqdn)
    ping_device(proxy)
    weak_ref = weakref.ref(proxy)

    str(proxy)

    # clear strong reference and check if object can be garbage collected
    del proxy
    assert_object_released_after_gc(weak_ref)


def test_no_cyclic_ref_for_proxy(green_mode_device_proxy, simple_device_fqdn):
    proxy = green_mode_device_proxy(simple_device_fqdn)
    ping_device(proxy)
    weak_ref = weakref.ref(proxy)

    # clear strong reference and check if object is immediately released
    del proxy
    assert_object_released_without_gc(weak_ref)


def assert_object_released_after_gc(weak_ref):
    gc.collect()
    assert weak_ref() is None


def assert_object_released_without_gc(weak_ref):
    try:
        assert weak_ref() is None
    except AssertionError:
        if weak_ref().get_green_mode() == GreenMode.Asyncio:
            pytest.xfail('Sometimes fails with a concurrent.Future ref cycle')
        else:
            raise
